/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.api;

/**
 * This is the interface for Concurrency Control service.
 *
 * @author Luciano Garcia-Banuelos (Luciano.Garcia@imag.fr)
 */
public interface ConcurrencyManager {

    /**
	 * This method records the start of an execution context. It can be a
	 * transaction starting.
	 * @param ctx is the context
	 */
	void begin(Object ctx);

    /**
	 * This method requests the validation of accesses made by the context. 
     * @return a boolean value to indicate if the accesses are validated or
	 * not.
	 */
	boolean validate(Object ctx);

    /**
	 * This method marks the end of accesses made by the execution context. This
	 * method should be called when the validate method has returned true.
	 */
	void finalize(Object ctx);

	/**
	 * This method allows to release the resources allocated in a given context.
	 * This method should be called when the validate method has returned false.
	 */
    void abort(Object ctx);

	/**
	 * This method records an access intention to a data object in read mode.
	 * @param ctx is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param resourceId is the resource identifier (Object used for 
	 * synchronization and equals)
	 * @param lockHints permit to take smaller lock than on the resource globaly
	 * @param resProv is used to retrieve the resource. If not resource provider
     * is specified (null value), then the method returns a null value.
	 * @param resourceHints is a hints helping the ResourceProvider to provide
     * the resource corresponding to the identier
	 * @return the resource obtained from the resource provider if there is one.
	 * @exception ConcurrencyException if there is a concurrency problem. In this
	 * last case that means the context should be cancelled.
	 */
    Object readIntention(Object ctx, 
            Object resourceId, 
            Object lockHints, 
            ResourceProvider resProv, 
            Object resourceHints) throws ConcurrencyException;

	/**
	 * This method records an access intention to a data object in write mode.
	 * A call to the readIntention is necessary before a call to this method.
	 * @param ctx is an identifier of the execution context. It can be a
	 * transaction handle.
     * @param resourceId is the resource identifier (Object used for 
     * synchronization and equals)
	 * @param lockHints permit to take smaller lock than on the resource globaly
	 * @param resProv is used to retrieve the resource. If not resource provider
     * is specified (null value), then the method returns a null value.
     * @param resourceHints is a hints helping the ResourceProvider to provide
     * the resource corresponding to the identier
	 * @return the resource obtained from the resource provider if there is one.
	 * @exception ConcurrencyException if there is a concurrency problem. In this
	 * last case that means the context should be cancelled.
	 */
    Object writeIntention(Object ctx, 
            Object resourceId, 
            Object lockHints, 
            ResourceProvider resProv, 
            Object resourceHints) throws ConcurrencyException;
    
    /**
     * This interfaces is the callback used by the concurrency manager to obtain
     * a resource from its identifier on readIntention or writeIntention calls.
     *
     * @author S.Chassande-Barrioz
     */
    interface ResourceProvider {

        /**
         * Retrieves the resource corresponding to an identifier.
         * @param context is the execution context (typically a working set).
         * @param resourceId is the identifier of the expected resource. It must be
         * non null.
         * @param resourceHints is a hints helping the ResourceProvider to provide
         * the resource corresponding to the identier
         * @param lockHints is a hint reducing conflict. It can be null. 
         * @param resourceCopyRequired indicates if the concurrency policy requires
         * a new copy of the resource (ex: copy on write)
         * @param suggestedResource is a resource instance suggested by the 
         * concurrency manager it self. Indeed when the concurrency manager has 
         * obtained the right for the resource, maybe the resource has been fetched
         * too. This parameter can be null if the concurrency manager has no 
         * resource instance to propose.
         * @param synchronizer is a java object to synchronize in order to avoid
         * concurrent loading between threads. 
         * @return a resource or a null value. This value will be returned by the
         * ConcurrencyManager
         * @throws ConcurrencyException if there is not resource corresponding to
         * the identifier.
         */
        Object getResource(Object ws, 
                Object resourceId,
                Object resourceHints,
                Object lockHints,
                boolean resourceCopyRequired,
                Object suggestedResource,
                Object synchronizer) throws ConcurrencyException;
    }
}
