package org.objectweb.perseus.concurrency.distributed;

/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@rd.francetelecom.com)
 *
 */

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.distributed.globallock.api.DeadLockException;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLock;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockException;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockManager;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockWaiter;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.concurrency.pessimistic.Lock;
import org.objectweb.perseus.concurrency.pessimistic.PessimisticConcurrencyManager;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Map;

public abstract class DistributedConcurrencyManager
        extends PessimisticConcurrencyManager implements AttributeController {

	public final static String GLOBAL_LOCK_MANAGER_BINDING = "global-lock-manager";


	private long timeout = 0;
	private GlobalLockManager gLockMgr;

	public void setTimeOut(long timeout) {
		this.timeout = timeout;
	}


	public DistributedConcurrencyManager() throws ConcurrencyException {
		super();
	}

	// BindingController interface
	public String[] listFc() {
		String[] super_res = super.listFc();
		String[] res = new String[super_res.length + 1];
		int cpt;
		for (cpt = 0; cpt < super_res.length; cpt++) {
			res[cpt] = super_res[cpt];
		}
		res[cpt] = GLOBAL_LOCK_MANAGER_BINDING;
		return res;
	}

	public Object lookupFc(String s) throws NoSuchInterfaceException {
		if (GLOBAL_LOCK_MANAGER_BINDING.equals(s)) {
			return gLockMgr;
		}
		return super.lookupFc(s);
	}

	public void bindFc(String s, Object o) throws IllegalBindingException,
			NoSuchInterfaceException {
		if (GLOBAL_LOCK_MANAGER_BINDING.equals(s)) {
			if (gLockMgr != null) {
				throw new IllegalBindingException(s + ": Already bound");
			}
			try {
				gLockMgr = (GlobalLockManager) o;
			} catch (ClassCastException e) {
				throw new IllegalBindingException(s + ":"  + e.getMessage());
			}
		} else super.bindFc(s, o);
	}

	public void unbindFc(String s) throws NoSuchInterfaceException {
		if (GLOBAL_LOCK_MANAGER_BINDING.equals(s)) {
			gLockMgr = null;
		} else super.unbindFc(s);
	}

	// ConcurrencyManager interface
	public void begin(Object ctx) {
		super.begin(ctx);
	}

	public void finalize(Object ctx) {
	    freeLocks(ctx);
	}
	private void freeLocks(Object ctx) {
		Object resourceId;
		synchronized (locks) {
			Iterator entries = locks.entrySet().iterator();
			while (entries.hasNext()) {
				Map.Entry entry = (Map.Entry) entries.next();
				Lock lock = (Lock) entry.getValue();
				resourceId  = entry.getKey();
				GlobalLock gl = null;
				try {
					gl = getDistLock(resourceId, lock.hints, true);
				} catch (ConcurrencyException e) {
					logger.log(BasicLevel.ERROR,
							"Error during the finalize of the context " + ctx
							+ ": oid=" + resourceId, e);
				}

				if (lock.close(ctx)) {
					gl.downgrade(RWLockValue.NOLOCK);
					entries.remove();
				} else {
					gl.downgrade(lock.getMax());
				}
				if (gl.getGrantable() == RWLockValue.NOLOCK)
					invalidateState(resourceId, lock.hints);
			}
		}
		contextInfos.remove(ctx);
	}

	public void abort(Object ctx) {
	    freeLocks(ctx);
	}

	private Object accessIntention(Object ctx, Object resourceId,
								   Object thinLock, byte lck, 
                                   ResourceProvider resProv,
                                   Object resourceHints)
			throws ConcurrencyException {
		try {
			Lock l = getLock(resourceId, thinLock);
			GlobalLock gl = getDistLock(resourceId, l.hints, true);
			GlobalLockWaiter w = gl.upgrade(lck, false, timeout);
			if (w != null) {
				if (w.waitLock(timeout) == false) {
					w.signalHandled();
					throw new DeadLockException();
				}
			}
			if (lck == RWLockValue.READ) { 
                l.readIntention(ctx);
            } else { 
                l.writeIntention(ctx);
            }
			if (w != null) {
                w.signalHandled();
            }
            if (resProv == null) {
                return null;
            } else {
                return resProv.getResource(ctx, resourceId, resourceHints, thinLock, false, null, l.oid);
            }
		} catch (ConcurrencyException e) {
			getContextInfo(ctx).rollback = true;
			throw e;
		} catch (Exception e) {
			throw new ConcurrencyException(e);
		}

	}
	public Object readIntention(Object ctx, Object resourceId, Object hints, 
            ResourceProvider resProv, Object resourceHints) throws ConcurrencyException {
		return accessIntention(ctx, resourceId, hints, RWLockValue.READ, resProv, resourceHints);
	}

	public Object writeIntention(Object ctx, Object resourceId, Object hints, 
            ResourceProvider resProv, Object resourceHints) throws ConcurrencyException {
		return accessIntention(ctx, resourceId, hints, RWLockValue.WRITE, resProv, resourceHints);
	}


	// specific interface
	public void uncacheGlobal(Object oid) {
		try {
			GlobalLock gl = gLockMgr.getGlobalLock((Serializable) oid , false);
			if (gl != null)
				gl.uncache();
		} catch (GlobalLockException e) {
		}
	}

	public boolean isDistGrantable(Object resId, byte lck, Object hints)
			throws ConcurrencyException {
		GlobalLock distLock = getDistLock(resId, hints, false);
		return (distLock.getGrantable() >= lck);
	}

	// should be controlled w/ AttributeController
	public long getTimeout() {
		return timeout;
	}

	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}

	// private/protected interface
	protected GlobalLock getDistLock(Object resourceId, Object hints,
										 boolean doInvalidate)
			throws ConcurrencyException {

		GlobalLock gl;
		try {
			gl = gLockMgr.getGlobalLock((Serializable) resourceId, true);
		} catch (GlobalLockException e) {
			throw new ConcurrencyException(e);
		}
		if (doInvalidate && (gl.getGrantable() == RWLockValue.NOLOCK)) {
			invalidateState(resourceId, hints);
		}
		return gl;
	}

	/**
     * Invalidate the reference state of the resource
     * @param resourceId
     * @param hints
     */
    protected abstract void invalidateState(Object resourceId, Object hints);

}
