/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@rd.francetelecom.com)
 * Date: 2 juin 2004
 * Time: 17:01:42
 */

package org.objectweb.perseus.concurrency.distributed.globallock.api;


/**
 * This is the interface of global locks. Global locks are created and
 * associated w/ resource identifiers using the <code>getGlobalLock</code>
 * of the <code>GlobalLockManager</code> interface. Once created the lock
 * can be upgraded/downgraded to a specified level. The initial value is
 * implementation dependent.
 * @see GlobalLockManager
 */

public interface GlobalLock {

	/**
	 * Request a lock upgrade. In synchronous mode the method returns when
	 * the lock has been granted. In asynchronous mode the method returns a
	 * waiter object if the request cannot be granted immediately. The caller
	 * must call the waitLock method of this object to effectively wait
	 * for the lock. It must also call the signalHandled method when the
	 * granting has been handled to let other waiters be unblocked. This scheme
	 * let the possibility to implement a FIFO handling of requests.
	 * @param lck the requested lock level
	 * @param sync if true the method blocks until the request is granted
	 * @param timeout specifies the the maximum time to wait in sync mode
	 * @return a waiter object in asynchronous mode (sync == false) if the
	 * request cannot be satisfied immediately.
	 * @throws DeadLockException if the timeout has expired
	 * @throws InterruptedException if the thread has been interrupted while waiting
	 * @see GlobalLockWaiter
	 */
	public GlobalLockWaiter upgrade(byte lck, boolean sync, long timeout)
			throws DeadLockException, InterruptedException;

	/**
	 * Notifies a lock downgrade.
	 * @param lck the new level wanted for the lock
	 */ 
	public void downgrade(byte lck);


	/**
	 * Get the maximum lock level that can be granted immediately without
	 * contacting the global lock coordinator (and so without blocking).
	 * Note that this is just a hint. There is no guarantee that a subsequent
	 * call to upgrade will not block.
	 * @return the maximum grantable lock level
	 */
	public byte getGrantable();

	/**
	 * Uncache the lock level. The following call to upgrade will automatically
	 * generate a request to the coordinator.
	 */
	public void uncache();

}
