/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR AE PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@rd.francetelecom.com)
 *
 */

package org.objectweb.perseus.concurrency.distributed.globallock.lib;

import org.objectweb.perseus.distribution.api.DistResCoordinator;
import org.objectweb.perseus.distribution.api.DistResCoordinatorService;
import org.objectweb.perseus.distribution.api.NotCoordinatorException;
import org.objectweb.perseus.concurrency.lib.LockValue;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.io.Serializable;

/**
 * This class is the server side of the global lock management and is not
 * intended to be used directly. Coordinators will be dynamically created by
 * the distributed respurce management through the GlobalLockCoordinatorFactory.
 * Coordinators communicate with users using the DistResCoordinatorService
 * interface. As global locks are cached by users, they must be called back
 * when necessary. The coordinator ensure that a lock is never granted to
 * different users with incompatible levels. Incompatibility is defined by the
 * lockValue when initializing the GlobalLockCoordinatorFactory instance. When
 * a user request cannot be satisfied because it is not compatible with other
 * users, a downgrade lock authoritative request (that is, a call back) is sent
 * users with a level compatible with the requested one. On reception, users
 * must adjust their grantable lock level so that new
 * local reuqest will be blocked, and notify the coordinator of the effective
 * downgrading when the local granted level is less or equal to the call back
 * level. Note that new coordinator instance is created by the
 * GlobalLockMasterFactory for each shared resource. Hence the instance itself
 * is sent when coordination migration or replication.
 * @see org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockUser
 * @see org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockCoordinatorFactory
 * @see org.objectweb.perseus.concurrency.lib.LockValue
 * @see org.objectweb.perseus.distribution.api.DistResCoordinator
 * @see org.objectweb.perseus.distribution.api.DistResCoordinatorService
 */

public class GlobalLockCoordinator implements DistResCoordinator, Serializable {

	static final boolean trace = true;
	// the id of the resource to which the locks are attached
	private Object objId;
	private HashSet users;	// the set of nodes that use this lock
	private byte maxGrantedLock;// maximum lock level granted to users

	private LinkedList waiting;		// the list of nodes waiting for callbacks
	private byte requestedDowngrade;// the level of the current call back request
	private int nbDowngradeRequests;// the number of nodes that do not have yet
									// answered to the call back request

	// transient fields that must not be serialized for coordination operations
	transient DistResCoordinatorService drcs; // the service to use to communicate
	transient LockValue lockValue;	// the lock value to use for lock comparison
	transient private int lastCBnumber;


	/**
	 * Create a new coordinator using a LockValue
	 * @param lockValue
	 * @see org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockCoordinator
	 * @see org.objectweb.perseus.concurrency.lib.LockValue
	 */

	public GlobalLockCoordinator(Object resId, DistResCoordinatorService drcs,
								 LockValue lockValue) {
		initCopy(resId, drcs, lockValue);
		users = new HashSet();
		waiting = new LinkedList();
		// no lock level has been granted for now
		maxGrantedLock = LockValue.NOLOCK;
		// there is no current call back requests
		requestedDowngrade = lockValue.maxValue();
		nbDowngradeRequests = 0;
		// TODO: using a random value make the probability not ignoring a obsolete
		// call back response very very low but not null. We should take into account
		// something like the node id
		lastCBnumber = (new Random()).nextInt();
	}

	public GlobalLockCoordinator(GlobalLockCoordinator glc) {
		this.objId = glc.objId;
		this.users = glc.users;
		this.maxGrantedLock = glc.maxGrantedLock;
		this.waiting = glc.waiting;
		this.requestedDowngrade = glc.requestedDowngrade;
		this.nbDowngradeRequests = glc.nbDowngradeRequests;
		this.drcs = glc.drcs;
		this.lockValue = glc.lockValue;
	}
	// DistResCoordinator interface
		synchronized public void receive(Object objId, Serializable user,
										 Serializable msg) {
		if (!objId.equals(this.objId)) throw new InternalError();
		if (msg instanceof GlobalLockMessage) {
			GlobalLockMessage req = (GlobalLockMessage) msg;
			if (req.type == GlobalLockMessage.UPGRADE_REQUEST)
				upgrade(user, req.lck);
			else if (req.type == GlobalLockMessage.DOWNGRADE_NOTIFY)
				downgrade(user, req.lck, req.serialNumber);
			else if (req.type == GlobalLockMessage.UPGRADE_CANCEL)
				cancelUpgrade(user);
			else throw new InternalError();
		} else throw new InternalError("Unexpected message type");
	}

	synchronized public Serializable freeze(Object resId) {
		if (trace) trace("Freeze !!");
		return this;
	}

/*	synchronized public DistResCoordinator
			joinCoordinators(Serializable resId, Object copy,
							 DistResCoordinatorService drcs) {
		GlobalLockCoordinator res;
		this.drcs = drcs;
		this.objId = resId;
		if (copy != null) {
			if (trace) trace("Get coordination w/ copy");
			res = (GlobalLockCoordinator) copy;
			res.lockValue = lockValue;
			res.joinCoordinators(resId, null, drcs);
			return res;
		}
		if (trace) trace("New coordination");
		return this;
	}
*/
	synchronized public boolean joinUsersRequest(Object resId,
										  Serializable node) {
		users.add(node);
		if (trace) trace("Accept user: " + node);
		return true;
	}

	public void recover(Object resId, Map userStates) {

		for (Iterator it = userStates.entrySet().iterator(); it.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			Serializable nodeId = (Serializable) entry.getKey();
			GlobalLockUser glu = (GlobalLockUser) entry.getValue();
			if (trace) trace("Received state from " + nodeId + ": " + glu);
			users.add(nodeId);
			if (maxGrantedLock < glu.locallyGranted) maxGrantedLock
					= glu.locallyGranted;
			if (glu.globallyRequested != LockValue.NOLOCK)
				waiting.addLast(new Waiter(nodeId, glu.globallyRequested));
		}
		if (!waiting.isEmpty()) {
			Waiter firstWaiting = (Waiter) waiting.get(0);
			callBackLocks(firstWaiting.nodeId, firstWaiting.lck);
		}
	}

	public void nodeFailed(Object resId, Serializable nodeId) {
		// remove pending requests of dead node
		for (Iterator it = waiting.iterator(); it.hasNext();) {
			Waiter waiter = (Waiter) it.next();
			if (waiter.nodeId.equals(nodeId))
				it.remove();
		}

	}

	// package methods

	// called by the factory
	void initCopy(Object resId, DistResCoordinatorService drcs,
				  LockValue lockValue) {
		this.objId = resId;
		this.drcs = drcs;
		this.lockValue = lockValue;

	}


	// private methods
	private void upgrade(Serializable n, byte lck) {
		try {
			if (trace) trace(
					"From: " + n + " GLOBAL REQUEST: " + lck);

			// treat only one request at a time in FIFO order
			if (requestedDowngrade != lockValue.maxValue()) {
				if (trace) System.out.println(this
						+ "From: " + n + " ALREADY PENDING REQUEST: " + lck);
				waiting.addLast(new Waiter(n, lck));
				if (trace) System.out.println(this
						+ "From: " + n + " GLOBAL REQUEST QUEUED (NOT FIRST): "
						+ lck);
				return;
			}

			// check if the max granted lock level is compatible
			// with the requested one
			if (lockValue.isCompatibleWith(lck, maxGrantedLock)) { // if so,
				// update the maxgranted lock
				if (maxGrantedLock < lck) maxGrantedLock = lck;

				// notify the user that the lock level is granted
				drcs.sendToUser(objId,
						new GlobalLockMessage(GlobalLockMessage.UPGRADE_NOTIFY,
						lck, (long) 0, null), n);
				return;
			}
// TODO: lockValue.isCompatibleWith(lck, maxGrantedLock) not redundant with previous ??
			// check if the requesting user is the sole user
			if (lockValue.isCompatibleWith(lck, maxGrantedLock) ||
					(users.size() == 1 && users.contains(n)) ) { // if so,
				// update the maxgranted lock: equal to what is requested
				// because the request may follow an uncache
				// operation on the user. So we set max GrantedLock to
				// lck even if it was greater than it
				maxGrantedLock = lck;

				// notify the user that the lock level is granted
				drcs.sendToUser(objId,
						new GlobalLockMessage(GlobalLockMessage.UPGRADE_NOTIFY,
								lck, (long) 0, null), n);
				return;
			}

			// call back a lock level compatible with the requested one
			byte downgradeLock = lockValue.getCompatibleWith(lockValue.maxValue(),
					lck);
			// queue request
			Waiter w= new Waiter(n, lck);
			waiting.addLast(w);
			callBackLocks(n, downgradeLock);

			if (trace) trace("From: "
					+ n + " GLOBAL REQUEST QUEUED (FIRST): " + lck);

		} catch (NotCoordinatorException e) {
			// this should not happen !!
			throw new InternalError("Not coordinator of this resource !!!");
		}
	}

	private void callBackLocks(Serializable n, byte downgradeLock) {
		try {

			// update current callback lock level
			requestedDowngrade = downgradeLock;

			// send a callback request to all users but the requesting one
			HashSet dest = new HashSet(users);
			dest.remove(n);
			nbDowngradeRequests = dest.size();
			if (trace) trace("From: "
				+ n + " SND CALLBACK: " + downgradeLock + " TO: " + dest);
			lastCBnumber++;
			drcs.sendToUsers(objId,
				new GlobalLockMessage(GlobalLockMessage.DOWNGRADE_REQUEST,
						downgradeLock, (long) 0, n, lastCBnumber), dest);

		} catch (NotCoordinatorException e) {
			// this should not happen !!
			throw new InternalError("Not coordinator of this resource !!!");
		}
	}

	private void downgrade(Object n, byte lck, int serialNumber) {
		try {
			// After a crash, the waiters' list may have been re-ordered
			// and new global lock downgrade requests may have been sent.
			// Hence, we may receive response related to callback requests
			// sent before the crash so we need to ignore them.
			// We use a serial number to distinguish responses. It it initialized
			// with a random value
			if (serialNumber != lastCBnumber) {
//			if ((waiting == null) || waiting.isEmpty() ||
//					(!pending.equals(((Waiter) waiting.getFirst()).nodeId))) {
				if (trace)
					trace("From: " + n + " GLOBAL RELEASE IGNORED " + lck);
				return;
			}
			// one less waiting request
			nbDowngradeRequests--;

			if (trace) trace("From: " + n + " GLOBAL RELEASE: " + lck);

			// if there still are waiting request, nothing else to do
			if (nbDowngradeRequests > 0)
				return;

			// max granted lock level now equals the requested level
			maxGrantedLock = requestedDowngrade;

			// no callback is currently processed
			requestedDowngrade = lockValue.maxValue();

			// for each waiting node, check that its request can be granted,
			// and if so, notify it
			for (Iterator it = waiting.iterator(); it.hasNext();) {
				Waiter w = (Waiter) it.next();
				if (lockValue.isCompatibleWith(w.lck, maxGrantedLock)) {
					if (maxGrantedLock < w.lck) maxGrantedLock = w.lck;
					it.remove();
					if (trace) trace("To: "
							+ w.nodeId + " NOTIFY GLOBAL UPGRADE: " + lck);
					drcs.sendToUser(objId,
							new GlobalLockMessage(GlobalLockMessage.UPGRADE_NOTIFY,
									w.lck, (long) 0, null),
							w.nodeId);
				} else {
					callBackLocks(w.nodeId,
							lockValue.getCompatibleWith(lockValue.maxValue(),
									w.lck));
					break;
				}
			}

		} catch (NotCoordinatorException e) {
			// this should not happen !!
			throw new InternalError("Not coordinator of this resource !!!");
		}
	}
	private void cancelUpgrade(Object n) {
		Waiter w = null;
		for (Iterator it = waiting.iterator(); it.hasNext();) {
			w = (Waiter) it.next();
			if (w.nodeId.equals(n))
				break;
		}
		if ((w == null) || !w.nodeId.equals(n)) {
			if (trace) trace("From: " + n + " UNEXPECTED CANCELATION !! (IGNORE IT)");
		}

		if (waiting.indexOf(w) == 0) {
			if (trace) trace("From: " + n + " CANCEL PENDING REQUEST (FIRST WAITING)");
			waiting.removeFirst();
			if (!waiting.isEmpty()) {
				w = (Waiter) waiting.getFirst();
				if (trace) trace("From: " + n + " CANCEL PENDING REQUEST (NEW CALL BACK)");
				callBackLocks(w.nodeId,
						lockValue.getCompatibleWith(lockValue.maxValue(),
						w.lck));
			} else {
				// no more downgrade requests
				requestedDowngrade = lockValue.maxValue();
				nbDowngradeRequests = 0;
				if (trace) trace("From: " + n + " CANCEL PENDING REQUEST (NO MORE PENDING)");
			}
		} else {
			waiting.remove(w);
			if (trace) trace("From: " + n + " CANCEL PENDING REQUEST (NOT FIRST WAITING)");
		}
	}
	// class used to queue waiting request
	private class Waiter implements Serializable {
		public Serializable nodeId;
		byte lck;
		//int serialNumber;
		Waiter(Serializable n, byte lck) {
			this.lck = lck;
			this.nodeId = n;
		}
	}


	public String toString() {
		return "=====MASTER OF " + objId
				+ "(" + drcs.getNodeId() + ")[" + maxGrantedLock
				+ requestedDowngrade + "(" + nbDowngradeRequests
				+ "/" + users.size() + ")] T="
				+ Thread.currentThread().hashCode() + " ";
	}

	private void trace(String s) {
		System.out.println(this + " " + s);
	}


}
