/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR AE PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@rd.francetelecom.com)
 *
 */

package org.objectweb.perseus.concurrency.distributed.globallock.lib;

import org.objectweb.perseus.distribution.api.DistResUser;
import org.objectweb.perseus.distribution.api.DistResUserService;
import org.objectweb.perseus.distribution.api.NotUserException;
import org.objectweb.perseus.concurrency.distributed.globallock.api.DeadLockException;
import org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockMessage;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLock;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockWaiter;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.concurrency.lib.LockValue;

import java.io.Serializable;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * This class the client side of global lock management. A global lock can be
 * associated to each resource that is to be shared between different
 * machines (actually JVMs).
 * Global lock users are coordinated by a GlobalLockCoordinator and
 * communicate with it through the DistResUserService interface.
 * Global locks permit to share a resource between nodes, but not between tasks.
 * One must ensure that the global lock that has been obtained locally
 * is sufficient to satisfy the requirements of the local tasks, and that
 * the local tasks are well synchronized according to the local lock levels they
 * have obtained. Locks are managed in term of level. The lock level must be
 * explicitly upgraded if what is grantable not sufficient to satisfy new local
 * requirements and must be explicitly downgraded to the maximum local
 * requirements when local requirements are released.
 * Locks are cached: the local node indicates what is locally required but
 * internally but what will be locally grantable can be bigger that what has
 * been actually granted. The local lock level can be downgraded to NOLOCK,
 * then later upgraded to some lock value without inducing a request to the
 * coordinator provided the grantable level is  bigger thant the new requested
 * level. Lock levels are called back by the coordinator when required.
 * For each shared resource the user state is composed of three lock level
 * 		- what HAS BEEN granted locally
 *      - what CAN BE granted locally
 *      - what has been requested to the coordinator (which is hidden to users)
 * Note that the different level of lock and their compatibility is configured
 * through the LockValue interface. This implementation works well with
 * Shared/Exclusive locks but should also work with more complex lock values
 * like intention to read/write lock levels, provided lock levels are strictly
 * ordered. This can be configured when initializing the GlobalLockMasterFactory
 * Note that at the moment, deadlocks are detected with timeout. AE timeout can
 * occur if the lock level cannot be granted because of its use, or because
 * there are some communication problems. However, when in the future fault
 * tolerence will be supported, the two behaviors will be distinguished. In
 * this case, the timeout value will be a hint to indicate a normal amount
 * of time a lock request should be blocked according to the application
 * behaviour. The time spent due to failure correction should not be taken into
 * account.
 * @see org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockCoordinator
 * @see org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockCoordinatorFactory
 * @see org.objectweb.perseus.concurrency.lib.LockValue
 * @see org.objectweb.perseus.distribution.api.DistResUser
 * @see org.objectweb.perseus.distribution.api.DistResUserService
 */

public class GlobalLockUser
		implements GlobalLock, DistResUser, Serializable {

	static final boolean trace = true;

	transient public DistResUserService drus;

	// name of user for tracing purpose (default is its nodeId)
	// should be removed at term
	// transient private String name;

	// object id to which is associated the lock
	transient Serializable objId;
	// max lock level that has been granted locally
	byte locallyGranted;
	// max lock level that can be granted locally without contacting the master
	byte locallyGrantable;
	// max lock level that has been requested to the master
	byte globallyRequested;

	// list of waiters
	transient private List waiters;

	// id of a global requester of a lock
	//transient private Object requester;
	transient int callBackSN;

	public GlobalLockUser(Serializable objId, DistResUserService drus) {
		common_init(objId, drus);
		// this.name = name;
	}

/*	public String getName() {
		// return name;
	}
*/

	private void createState(){
	//	state = new State();
		locallyGrantable = LockValue.NOLOCK;
		locallyGranted = LockValue.NOLOCK;
		globallyRequested = LockValue.NOLOCK;
		// to limit memory allocation, waiters object could be allocated only
		// when there are waiters
		waiters =  new LinkedList();

	}
	private void common_init(Serializable objId, DistResUserService drus) {
		this.objId = objId;
		//if (this.drus == null)
			this.drus = drus;
		//else if (this.drus != drus)
		//	throw new IllegalArgumentException("Cannot handle multiple " +
		//			"DistResUserService");
		createState();
	}

	/* local interfaces */
	synchronized public byte getGrantable() {
		return locallyGrantable;
	}


	public GlobalLockWaiter upgrade(byte lck, boolean sync, long timeout)
			throws DeadLockException, InterruptedException {
		GlobalLockWaiterImpl w;
		synchronized(this) {
			if (trace) trace("LOCAL REQUEST: " + lck);

			// check if the lock is grantable
			if (lck <= locallyGrantable) {
				// adjust what has been granted
				if (lck > locallyGranted) locallyGranted = lck;
				if (trace) trace("LOCAL REQUEST GRANTED: " + lck);
				return null;
			}
			w = waitForGrantable(lck, timeout);
		}

		if (sync) {
			boolean ok = w.waitLock(timeout);
			if (!ok) {
					if (trace) trace("WAKE UP TIMED OUT (Waiting: " + lck + ")");
					throw new DeadLockException();
			}

			if (trace) trace("WAKE UP (Waiting: " + lck + ")");
			w.signalHandled();
			return null;
		} else {
			return w;
		}
	}


	synchronized public void downgrade(byte lck) {
		if (trace) trace("LOCAL DOWNGRADE: " + lck);
		//if (state == null) createState();
		// if grantable is less than granted, this means that a call-back
		// request has been received from the coord for a lock that was locally
		// granted at that time. Hence we must satisfy the request now that we
		// can do it
		boolean replyCB = ((locallyGrantable < locallyGranted)
				&& (lck <= locallyGrantable));
		if (locallyGranted > lck) {
			locallyGranted = lck;
			if (replyCB) {
				if (trace) trace("CALL BACK RESP: " + lck + " SN=" + callBackSN);
				sendCoordinator(
						new GlobalLockMessage(GlobalLockMessage.DOWNGRADE_NOTIFY,
								lck, (long) 0, null, callBackSN));
				callBackSN = -1;
			}
		}

		// cancel pending request if not valid anymore
		// this happens when aborting a tx (e.g. after a dealock) which was
		// waiting for a global lock to be granted by the coord
		// Take care that there may be other threads waiting
		if (globallyRequested > LockValue.NOLOCK) {
			byte l = LockValue.NOLOCK;
			for (Iterator it = waiters.iterator(); it.hasNext();) {
				GlobalLockWaiterImpl waiter = (GlobalLockWaiterImpl) it.next();
				byte wl = waiter.getLockLevel();
				if (wl > l) l = wl;
			}
			if (l == LockValue.NOLOCK) {
				if (trace) trace("CANCEL PENDING REQUEST " + lck);
				sendCoordinator(new GlobalLockMessage(GlobalLockMessage.UPGRADE_CANCEL,
									lck, (long) 0, null));
			}
			globallyRequested = l;
		}
	}


	synchronized public void uncache() {
		if (trace) trace("UNCACHE");
		locallyGrantable = LockValue.NOLOCK;
	}

	// private methods called on message reception

	// called when the coord wants to call a lock back
	synchronized private void globalDowngrade(byte lck, Object requester, int SN) {
		if (trace) trace("RCV GLOBAL DOWNGRADE (CB): " + lck + " for " + requester);
		if (locallyGrantable > lck) locallyGrantable = lck;
		if (lck >= locallyGranted) {
			if (trace) trace("SEND GLOBAL DOWNGRADE OK: " + lck);
			sendCoordinator(new GlobalLockMessage(GlobalLockMessage.DOWNGRADE_NOTIFY,
					lck, (long) 0, requester, SN));
		return;
		}
		//this.requester = requester;
		this.callBackSN = SN;
		if (trace) trace("SEND GLOBAL DOWNGRADE NOT OK: " + lck);
//TODO: notifies local that a downgrade is requested
		// callbackListner.notifyCallback(lck);
	}


	// called when the coord notifies a upgrade of the local lock
	synchronized private void globalUpgrade(byte lck) {
		GlobalLockWaiterImpl prevWaiter = null;
		if (trace) trace("RCV GLOBAL UPGRADE: " + lck);
		if (locallyGrantable < lck) locallyGrantable = lck;
		if (globallyRequested == lck) globallyRequested = RWLockValue.NOLOCK;

		if (trace) trace("WAKEUP waiters: " + lck);
		for (Iterator it = waiters.iterator(); it.hasNext();) {
			GlobalLockWaiterImpl waiter = (GlobalLockWaiterImpl) it.next();
			if (trace) trace("SIGNAL LOCK TO WAITER: " + waiter);
			// remember value of waiting lock because signalLock will modify it
			byte wLock = waiter.getLockLevel();
			if (!waiter.signalLock(lck, prevWaiter)) break;
			if (locallyGranted < wLock) locallyGranted = wLock;
			if (trace) trace("SIGNAL OK: " + waiter);
			prevWaiter = waiter;
			it.remove();
		}
	}


	private GlobalLockWaiterImpl waitForGrantable(byte lck, long timeout) {

		// send a request to the coord only if a not stronger enough request
		// has been already sent
		if (lck > globallyRequested) {
			globallyRequested = lck;
			if (trace) trace("SEND GLOBAL REQUEST: " + lck);
			sendCoordinator(
					new GlobalLockMessage(GlobalLockMessage.UPGRADE_REQUEST,
							lck, timeout, drus.getNodeId()));
		} else
			if (trace) trace("ALREADY AE PENDING REQUEST : " + globallyRequested);

		if (trace) trace(" BLOCKED (Waiting: " + lck + " for " + timeout
				+ " millis)");
		// create and insert the waiter object
		GlobalLockWaiterImpl w = new GlobalLockWaiterImpl(lck);
		if (trace) trace(" ADD WAITER: " + w);
		waiters.add(w);
		return w;
	}

	protected void sendCoordinator(Serializable message) {
		try {
			Serializable resId = objId;
			drus.sendToCoordinator(resId, message);
		} catch (NotUserException e) {
			e.printStackTrace();
		}

	}

	// called on reception of a message from the coordinator
	public void receive(Object objId, Serializable message) {
//		if (!objId.equals(this.objId)) throw new InternalError();
		if (message instanceof GlobalLockMessage) {
			GlobalLockMessage req = (GlobalLockMessage) message;
			if (req.type == GlobalLockMessage.UPGRADE_NOTIFY) {
				globalUpgrade(req.lck);
				// check for attached downgrade request
				//TODO: should use generic (lockValue) value here
				/* if ((req.auxLck != LockValue.UNDEFINED)
					&& (req.auxLck != RWLockValue.WRITE))
					globalDowngrade(req.auxLck, req.requester);
					*/
			} else if (req.type == GlobalLockMessage.DOWNGRADE_REQUEST)
				globalDowngrade(req.lck, req.requester, req.serialNumber);
			else throw new InternalError();
		} else throw new InternalError();
	}

	public synchronized Serializable getState(Object resId) {
		if (trace) trace("COORD SEEMS TO BE DOWN: return state");
		return this;
	}

	public synchronized String toString() {
		// handle case when drus is null (this can happen
		// during serialization)
		Object nodeId = (drus == null) ? null : drus.getNodeId();
			return "USER OF " + objId + " ("+ nodeId + ")["
					+ locallyGrantable + locallyGranted + globallyRequested
					+"] T=" + Thread.currentThread().hashCode() + " ";
	}

	protected void trace(String s) {
		System.out.println(this + " " + s);
	}

	void finilaze() {
		//TODO: should leave user's group
		//drus.leaveUsers(objId);
	}
}

