/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.dependency.api.DependencyGraph;

/**
 * A lock associated to an oid (see the "locks" map within the pessimistic
 * concurrency manager).
 * Provides a mutex concurrency policy.
 * @author E. Bruneton, P. Dechamboux
 */
public final class MutexLock extends Lock {
	/**
	 * The contexts that have a lock for the oid to which this object is
	 * associated.
	 */
	protected Object writer = null;
	/**
	 * The number of threads that wait for aquiring this lock.
	 */
	int waiter = 0;

	public MutexLock() {
	}

	public MutexLock(Object hints, DependencyGraph dg) {
		super(hints, dg);
	}

	/**
	 * Acquires this lock in read mode for the given context. This method
	 * blocks until the lock can be acquired in read mode by this context.
	 * @param ctxt a context.
	 */
	public void readIntention(Object ctxt)
			throws ConcurrencyException {
		writeIntention(ctxt);
	}

	/**
	 * Acquires this lock in write mode for the given context. This method
	 * blocks until the lock can be acquired in write mode by this context.
	 * @param ctxt a context.
	 */
	public synchronized void writeIntention(Object ctxt)
			throws ConcurrencyException {
		boolean ok;
		do {
			ok = ((writer == null) || writer.equals(ctxt));
			if (!ok) {
				waiter ++;
				try {
					wait();
				} catch (InterruptedException e) {
					throw new ConcurrencyException(
						"Waiting of a read intention has been interupted: ", e);
				} finally {
					waiter --;
				}
			}
		} while (!ok);
		reservations--;
		writer = ctxt;
	}

	/**
	 * Removes the given context from the reader and writer lists of this
	 * lock.
	 * @param ctxt a context
	 * @return true if the reader and writer list are empty, after the
	 * context has been removed from these lists. In such a case, this
	 * object can be removed from the 'locks' map.
	 */
	public synchronized boolean close(Object ctxt) {
		//TODO: wake up blocking threads working in the closed context
		if ((writer != null) && writer.equals(ctxt)) {
			writer = null;
		}
		boolean res = reservations == 0 && writer == null && waiter == 0;
		if (!res) {
			notifyAll();
		}
		return res;
	}

	public synchronized byte getMax() {
		if (writer != null) return RWLockValue.WRITE;
		return RWLockValue.NOLOCK;
	}
}
