/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.concurrency.lib.ReadWriteLockValue;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * @author E. Bruneton
 */
public class PessimisticConcurrencyManager
        implements ConcurrencyManager, BindingController,
		PessimisticConcurrencyManagerAC, LifeCycleController {

	public final static short POLICY_UNDEFINED = 0;
	public final static short POLICY_MUTEX = 1;
	public final static short POLICY_RW_PRIORITY2READERS = 2;
	public final static short POLICY_RW_FIFO = 3;
	public final static String[] policyStringVal = {
		"POLICY_UNDEFINED",
		"POLICY_MUTEX",
		"POLICY_RW_PRIORITY2READERS",
		"POLICY_RW_FIFO"
	};
	public final static String DEPENDENCY_GRAPH_BINDING = "dependency-graph";

	private String state = LifeCycleController.STOPPED;

	/**
	 * True to distinguish read/write locks, false to use a single lock for
	 * both modes.
	 */
	private short policy = POLICY_RW_FIFO;
	private final static short POLICY_LAST = POLICY_RW_FIFO;
	/**
	 * A map associating locks to all oids in all active contexts.
	 */
	protected Map locks;

    protected DependencyGraph dg = null;

	/**
	 * A map associating a ContextInfo to each context.
	 */
	protected Map contextInfos;

    protected Logger logger = null;
    
    protected boolean thinLockAllowed = true;

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
        return new String[] {DEPENDENCY_GRAPH_BINDING};
    }

    public Object lookupFc(String s) throws NoSuchInterfaceException {
        if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
            return dg;
        }
        throw new NoSuchInterfaceException(s);
    }

	public void bindFc(String s, Object o) throws IllegalBindingException,
			NoSuchInterfaceException {
		try {
			if ("logger".equals(s)) {
				logger = (Logger) o;
			} else if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
				if (dg != null) {
					throw new IllegalBindingException(s + ": Already bound");
				}
				dg = (DependencyGraph) o;
			} else {
                throw new NoSuchInterfaceException(s);
            }
		} catch (ClassCastException e) {
			throw new IllegalBindingException(s + ":"  + e.getMessage());
		}
	}

    public void unbindFc(String s) throws NoSuchInterfaceException {
		if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			dg = null;
		} else {
			throw new NoSuchInterfaceException(s);
		}
    }

	//IMPLEMENTATION OF THE ConcurrencyManager INTERFACE //
	//---------------------------------------------------//
	public String getFcState() {
		return state;
	}

	public void startFc() throws IllegalLifeCycleException {
		if (state == LifeCycleController.STARTED) return;
		locks = new HashMap();
		contextInfos = new HashMap();
		state = LifeCycleController.STARTED;
		switch (policy) {
		case POLICY_MUTEX:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager runs with POLICY_MUTEX.");
			break;
		case POLICY_RW_PRIORITY2READERS:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager runs with POLICY_RW_PRIORITY2READERS.");
			break;
		case POLICY_RW_FIFO:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager runs with POLICY_RW_FIFO.");
			break;
		}
	}

	public void stopFc() throws IllegalLifeCycleException {
		if (state == LifeCycleController.STOPPED) return;
		contextInfos = null;
		locks = null;
		state = LifeCycleController.STOPPED;
	}

	//IMPLEMENTATION OF THE ConcurrencyManager INTERFACE //
	//---------------------------------------------------//

	public void begin(Object ctx) {
		getContextInfo(ctx).rollback = false;
	}

	public boolean validate(Object ctx) {
		ContextInfo info = (ContextInfo) contextInfos.get(ctx);
		return info == null || !info.rollback;
	}

	public void finalize(Object ctx) {
	    freeLocks(ctx);
	}	    
	private void freeLocks(Object ctx) {
	    boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
	    if (debug) {
	        logger.log(BasicLevel.DEBUG, "Finalize the context: " + ctx);
	    }
		synchronized (locks) {
			Iterator entries = locks.entrySet().iterator();
			while (entries.hasNext()) {
				Map.Entry entry = (Map.Entry) entries.next();
				Lock lock = (Lock) entry.getValue();
				if (lock.close(ctx)) {
				    closeLock(lock, ctx);
					entries.remove();
				}
			}
		}
		contextInfos.remove(ctx);
	}

	public void closeLock(Lock lock, Object ctx) {
	}
	
	public void abort(Object ctx) {
	    freeLocks(ctx);
	}

	public Object readIntention(Object ctx, Object resourceId, Object thinLock, ResourceProvider resProv, Object resourceHints)
			throws ConcurrencyException {
		Lock lock = getLock(resourceId, null);
        try {
            lock.readIntention(ctx);
        } catch (ConcurrencyException e) {
            getContextInfo(ctx).rollback = true;
            throw e;
        }
        if (resProv == null) {
            return null;
        } else {
            try {
                return resProv.getResource(ctx, resourceId, resourceHints, thinLock, false, null, lock.oid);
            } catch (ConcurrencyException e) {
                removeUnavailableResource(ctx, lock);
                throw e;
            }
        }
	}

	public Object writeIntention(Object ctx, Object resourceId, Object thinLock, ResourceProvider resProv, Object resourceHints)
			throws ConcurrencyException {
	    if (thinLockAllowed && thinLock != null) {
	        // The application ask a thin lock on a resource. Firstly take 
	        // the read lock on resource in order to avoid concurrent 
	        // writers (resource lock vs thin lock)
			Lock lock = getLock(resourceId, null);
	        try {
	            lock.readIntention(ctx);
	        } catch (ConcurrencyException e) {
	            getContextInfo(ctx).rollback = true;
	            throw e;
	        }
	        if (lock.getMax() >= ReadWriteLockValue.UPGRADE) {
	            // The lock has already been taken with write intention 
	            // (upgrade or write) on the resource. Then a thin lock is
	            // useless.
                if (resProv == null) {
                    return null;
                } else {
                    try {
                        return resProv.getResource(ctx, resourceId, resourceHints, null, false, null, lock.oid);
                    } catch (ConcurrencyException e) {
                        removeUnavailableResource(ctx, lock);
                        throw e;
                    }
                }
	        } 
	        // The resouce lock level is not enough, then take the thin lock 
	        if (thinLock instanceof Collection) {
	            for(Iterator it = ((Collection) thinLock).iterator(); it.hasNext();) {
	                lock = getLock(resourceId, it.next());
	                try {
	                    lock.writeIntention(ctx);
	                } catch (ConcurrencyException e) {
	                    getContextInfo(ctx).rollback = true;
	                    throw e;
	                }
	            }
	            //As a read resource lock has been previously taken, pass only
	            // the last lock of the collection
                if (resProv == null) {
                    return null;
                } else {
                    try {
                        return resProv.getResource(ctx, resourceId, resourceHints, thinLock, false, null, lock.oid);
                    } catch (ConcurrencyException e) {
                        removeUnavailableResource(ctx, lock);
                        throw e;
                    }
                }
	        }
            logger.log(BasicLevel.WARN,
                    "Thin locking is not applicable with this lock hints: "
                    + thinLock + "\nNormal locking is used."); 
	    }
        Lock lock = getLock(resourceId, (thinLockAllowed ? thinLock : null));
        try {
            lock.writeIntention(ctx);
        } catch (ConcurrencyException e) {
            getContextInfo(ctx).rollback = true;
            throw e;
        }
        if (resProv == null) {
            return null;
        } else {
            try {
                return resProv.getResource(ctx, resourceId, resourceHints, thinLock, false, null, lock.oid);
            } catch (ConcurrencyException e) {
                removeUnavailableResource(ctx, lock);
                throw e;
            }
        }
	}

	//OTHER METHODS //
	//--------------//
    private void removeUnavailableResource(Object ctx, Lock lock) {
        synchronized (locks) {
            if (lock.close(ctx)) {
                locks.remove(lock);
            }
        }
    }

    private static class OID {
	    public Object oid;
	    public Object lockHints;
	    public OID(Object _oid, Object _lockHints) {
	        this.oid = _oid;
	        this.lockHints = _lockHints;
	        
	    }
	    public boolean equals(Object o) {
	        return o instanceof OID 
        	&& ((OID) o).oid.equals(oid)
    	    && ((OID) o).lockHints.equals(lockHints);
	    }
	}
	/**
	 * Returns the lock associated to the given oid.
	 * @param oid an object identifier.
	 * @return the lock associated to the given oid.
	 */
	protected Lock getLock(Object oid, Object hints) {
		if (thinLockAllowed && hints != null) {
		    oid = new OID(oid, hints);
		}
	    synchronized (locks) {
			Lock lock = (Lock) locks.get(oid);
			if (lock == null) {
				switch (policy) {
				case POLICY_MUTEX:
					lock = new MutexLock(hints, dg);
					break;
				case POLICY_RW_PRIORITY2READERS:
					lock = new RWPri2RLock(hints, dg);
					break;
				case POLICY_RW_FIFO:
					lock = new RWFifoLock(hints, dg);
					break;
				}
				lock.oid = oid;
				locks.put(oid, lock);
				if (logger != null) {
				    lock.logger = logger;
				    if (logger.isLoggable(BasicLevel.DEBUG)) {
				        logger.log(BasicLevel.DEBUG, "Allocate a Lock for the oid: " + oid);
				    }
				}
			}
			lock.reserve();
			return lock;
		}
	}

	protected ContextInfo getContextInfo(Object ctxt) {
		ContextInfo info = (ContextInfo) contextInfos.get(ctxt);
		if (info == null) {
			info = new ContextInfo();
			contextInfos.put(ctxt, info);
		}
		return info;
	}

	public String getPolicy() {
		return policyStringVal[policy];
	}

	public void setPolicy(String policy) throws IllegalLifeCycleException {
		if (!LifeCycleController.STOPPED.equals(state)) {
			throw new IllegalLifeCycleException("Cannot change lock policy of started pessimistic component");
		}
		for (short p = 0; p <= POLICY_LAST; p++) {
			if (policyStringVal[p].equals(policy)) {
				this.policy = p;
				return;
			}
		}
		throw new IllegalArgumentException("unrecognized locking policy: " + policy);
	}

	public boolean getThinkLockAllowed() {
	    return thinLockAllowed;
	}
	
	public void setThinkLockAllowed(boolean val) {
	    thinLockAllowed = val;
	}


	/**
	 * Information associated to a context
	 */
	protected static class ContextInfo {
		// long date; // date of begin()
		public boolean rollback; // if the validate() method must return false
	}
}
