/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.RolledBackConcurrencyException;
import org.objectweb.perseus.concurrency.lib.LockValue;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.concurrency.lib.ReadWriteLockValue;
import org.objectweb.perseus.concurrency.lib.Semaphore;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;

/**
 * A lock associated to an oid (see the "locks" map within the pessimistic
 * concurrency manager).
 * Provides "one writer/multiple readers" concurrency policy with a FIFO
 * scheduling.
 * @author E. Bruneton, P. Dechamboux, S.Chassande-Barrioz
 */
public final class RWFifoLock extends Lock {
    
	protected final static LockValue LOCKS = new ReadWriteLockValue();
	
    /**
     * The current active tasks which can access to the resource 
     */
	List owners = new ArrayList();
	
	/**
	 * The LockRequest instances that wait for lock for the oid to which this 
	 * object is associated.
	 */
	List waiters = new ArrayList();
	
	byte lockLevel = ReadWriteLockValue.NOLOCK;
	
	protected final Semaphore semaphore = new Semaphore();

	public RWFifoLock() {
	}

	public RWFifoLock(Object hints, DependencyGraph dg) {
		super(hints, dg);
	}

	/**
	 * @param task is the task identifier
	 * @param askedLock is the lock level asked by the task
	 * @param waiterIdx is the position of the lock request among waiters
	 * (0 means the request is the first)
	 * @return true if the lock request can be accepted.
	 */
	private final boolean acceptLockRequest(final Object task, 
	        final byte askedLock, final int waiterIdx) {
	    if (owners.isEmpty()) {
	        return true;
	    } else if (owners.contains(task)) {
	        //already has a lock
	        if (LOCKS.isCompatibleWith(askedLock, lockLevel)) {
	            //current lock is compatible
	            return waiterIdx == 0; // no waiter
	        } else if (owners.size() == 1) {
	            //only the task has the lock ==> up to to max level
	            //we do not take account of waiters, in order to avoid dead lock
	            //NOTE: this case happens when 
	            // - a task asks again a non shareable lock
	            // - OR a task increase it lock level
		        return true;
	        }
        } else if (LOCKS.isCompatibleWith(askedLock, lockLevel) 
	            && waiterIdx == 0) {
	        return true;
	    }
        return false;
	}
	
	/**
	 * Try to acquire a lock level for a given task.
	 *  
	 * @param task is the task identifier
	 * @param askedLock is the lock level asked by the task
	 * @throws RolledBackConcurrencyException if the lock request creates a 
	 * dea lock, or if the thread is intererupted during the waiting step.
	 */
	private final void accessIntention(final Object task, final byte askedLock)
			throws RolledBackConcurrencyException {
	    try {
        boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG); 
	    semaphore.P();
	    final int ws = waiters.size(); 
		reservations --;
		if (acceptLockRequest(task, askedLock, ws)) {
		    if (debug) {
		        logger.log(BasicLevel.DEBUG, LOCKS.str(askedLock) + " lock accepted directly,"
		                + "\n\ttask:" + task + "\n\towners=" + owners);
		    }
    	    this.lockLevel = (byte) Math.max(this.lockLevel, askedLock);
    	    if (!owners.contains(task)) {
    	        owners.add(task);
    	    }
		    semaphore.V();
		    return;
        }
		//must wait to obtain the lock level requested
		
	    //Compute dependencies
		List dependencies; // The dependency of the task
	    if (ws == 0) {
	        //no waiter, it depends to the owner,
	        dependencies = owners;
	    } else {
	        //There are waiters
	        LockRequest tmp = null;
	        int i;
	        //skip compatible requests
	        for(i = ws -1; i >= 0; i--) {
	            tmp = (LockRequest) waiters.get(i);
	            if (LOCKS.isCompatibleWith(askedLock, tmp.getLockLevel())) {
	            } else {
	                break;
	            }
	        }
	        if (i>=0) {
	            //dependencies is the list of compatible requests after 
	            // the skipped requests
		        dependencies = new ArrayList();
		        byte l = tmp.getLockLevel();
		        do {
		            tmp = (LockRequest) waiters.get(i);
			        dependencies.add(tmp.task);
		            if (!LOCKS.isCompatibleWith(tmp.getLockLevel(), l)) {
		                break;
		            }
		            i--;
		        } while (i>=0);
	        } else {
	            // all first are compatible, ==> dependencies are the owners
		        dependencies = owners;
	        }
	    }
	    //add a vertex for each dependency
	    if (dg.addVertexes(task, dependencies) != -1) {
		    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
		        logger.log(BasicLevel.INFO, LOCKS.str(askedLock) 
		                + " request: dead lock detected, cancel the task," 
		                + "\n\ttask:" + task);
		    }
		    //remove dependencies which can be added before the dead lock
	        dg.removeVertexes(task, dependencies);
	        //TODO: we could remove also the locks or requests of the same task
	        semaphore.V();
		    throw new RolledBackConcurrencyException("Deadlock");
	    }
	    if (debug) {
	        
	    }
	    //Create LockRequest and add it among waiters
	    LockRequest lockRequest = new LockRequest(askedLock, task);
        waiters.add(lockRequest);
	    
	    if (debug) {
	        StringBuffer sb = new StringBuffer();
	        sb.append(LOCKS.str(lockRequest.getLockLevel()));
	        sb.append(" lock waiting, rank=");
	        sb.append(waiters.indexOf(lockRequest));
	        sb.append(", task=");
	        sb.append(task);
	        sb.append("\n\towners=");
	        sb.append(owners);
	        printDG(sb);
	        logger.log(BasicLevel.DEBUG, sb.toString());
	    }
	    // wait
	    try {
		    synchronized(lockRequest) {
			    semaphore.V();
		        lockRequest.wait();
		    }
	        if (lockRequest.hasRolledBack()) {
	            forgetWaiter(lockRequest);
	            throw new RolledBackConcurrencyException(
					"The working set has been rolled back by another thread");
	        }
		} catch (InterruptedException e) {
		    forgetWaiter(lockRequest);
			throw new RolledBackConcurrencyException("Waiting of a " 
			        + LOCKS.str(lockRequest.getLockLevel()) 
			        + " intention has been interupted:", e);
		}
	    if (debug) {
	        logger.log(BasicLevel.DEBUG, LOCKS.str(lockRequest.getLockLevel()) 
	                + " Wake up\n\ttask:" + task);
	    }
	    } catch(Error t) {
	        logger.log(BasicLevel.ERROR, "accessIntention: " + t.getMessage(),t);
	        t.printStackTrace();
	    }

	}
	
	/**
	 * Eliminates a lock request among waiter. The dependencies are removed.
	 * @param lr is the lock request to forget
	 */
	private final void forgetWaiter(final LockRequest lr) {
        boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG); 
	    semaphore.P();
	    int idx = waiters.indexOf(lr);
	    if (debug) {
	        logger.log(BasicLevel.DEBUG, "forgetLockRequest: lock: "
	                + LOCKS.str(lr.getLockLevel()) 
	                + ", waiter rank=" + idx + "/" + waiters.size() 
	                + ", task=" + lr.task);
	    }
	    if (idx == -1 || waiters.size() == 0) {
		    semaphore.V();
	        return;
	    }
	    waiters.remove(idx);
	    
	    //Remove dependencies of the request to forget
        int i;
        LockRequest tmp = null;
        for(i = idx -1; i>=0; i--) {
	        tmp = (LockRequest) waiters.get(i);
	        if (!LOCKS.isCompatibleWith(
	                lr.getLockLevel(), 
	                tmp.getLockLevel())) {
	            break;
	        }
	    }
        if (i>=0) {
            byte l = tmp.getLockLevel();
            do {
                tmp = (LockRequest) waiters.get(i);
                dg.removeVertex(lr.task, tmp.task);
                if (!LOCKS.isCompatibleWith(tmp.getLockLevel(), l)) {
                    break;
                }
                i--;
            } while (i>=0);
        } else {
		    dg.removeVertexes(lr.task, owners);
        }
	    
	    //Remove dependencies to the request to forget
	    final int ws = waiters.size();
	    //skip compatible request
	    for(i = idx; i<ws; i++) {
	        tmp = (LockRequest) waiters.get(i);
	        if (!LOCKS.isCompatibleWith(
	                tmp.getLockLevel(), 
	                lr.getLockLevel())) {
	            break;
	        }
	    }
	    if (i<ws) {
	        byte l = tmp.getLockLevel();
	        do {
	            tmp = (LockRequest) waiters.get(i);
	            dg.removeVertex(tmp.task, lr.task);
	            if (!LOCKS.isCompatibleWith(tmp.getLockLevel(), l)) {
	                break;
	            }
	            i++;
	        } while (i<ws);
	    }
	    semaphore.V();
	}

	/**
	 * Acquires this lock in read mode for the given context. This method
	 * blocks until the lock can be acquired in read mode by this context.
	 * @param task a context.
	 */
	public void readIntention(Object task)
			throws ConcurrencyException {
	    accessIntention(task, RWLockValue.READ);
	}
	
	/**
	 * Acquires this lock in write mode for the given context. This method
	 * blocks until the lock can be acquired in write mode by this context.
	 * @param task a context.
	 */
	public void writeIntention(Object task)
			throws ConcurrencyException {
	    accessIntention(task, ReadWriteLockValue.WRITE);
	}
	
	/**
	 * Removes the given context from the reader and writer lists of this
	 * lock.
	 * @param task a context
	 * @return true if the reader and writer list are empty, after the
	 * context has been removed from these lists. In such a case, this
	 * object can be removed from the 'locks' map.
	 */
	public boolean close(Object task) {
	    try {
	    semaphore.P();
	    boolean isowner = owners.remove(task);
	    if (isowner) {
	        byte tmpLock = ReadWriteLockValue.NOLOCK;
	        //remove depedencies to the ended task
	        for(int i=0; i<waiters.size();i++) {
	            LockRequest lr = (LockRequest) waiters.get(i);
		        dg.removeVertex(lr.task, task);
	            if (LOCKS.isCompatibleWith(lr.getLockLevel(), tmpLock)) {
			        tmpLock = (byte) Math.max(lr.getLockLevel(), tmpLock);
	            } else {
	                break;
	            }
	        }
		    if (owners.isEmpty()) {
		        lockLevel = ReadWriteLockValue.NOLOCK;
		    }
	    } else {
	        //This case occurs when dead lock has been detected or when the 
	        // thread has been intereupted. In both case the fail lock request 
	        // has been cleaned (@see #accessIntention method).
	        // However if tasks are multithreaded (several thread with
	        // the same task identifier), it can stay lock request among the 
	        // waiters. This case is not yet managed
	    }
	    //try to notify waiters if possible
        int nbNotified = 0;
        while(!waiters.isEmpty()) {
            LockRequest lr = (LockRequest) waiters.get(0);
            if (!acceptLockRequest(lr.task, lr.getLockLevel(), 0)) {
                break;
            }
            waiters.remove(0);
	        dg.removeVertexes(lr.task, owners);	
	        if (!owners.contains(lr.task)) {
	            owners.add(lr.task);
	        }
            lockLevel = (byte) Math.max(lr.getLockLevel(), lockLevel);
            synchronized(lr) {
                lr.notify();
            }
            nbNotified++;
        }	    
	    if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
	        StringBuffer sb = new StringBuffer();
	        if (isowner) {
	            sb.append("task closed: ");
	        } else {
	            sb.append("task closed without lock: ");
	        }
	        sb.append(task);
	        sb.append(" / new notified: ");
	        sb.append(nbNotified);
	        sb.append(" / reservations:");
	        sb.append(reservations);
	        sb.append(" / current lock: ");
	        sb.append(LOCKS.str(lockLevel));
	        sb.append("\n\towners:");
	        sb.append(owners);
	        sb.append("\n\twaiters:");
	        sb.append(waiters);
	        printDG(sb);
	        logger.log(BasicLevel.DEBUG,  sb.toString());
	    }
	    try {
	        return reservations == 0 && owners.isEmpty();
	    } finally {
		    semaphore.V();
	    }
	    } catch(Error t) {
	        logger.log(BasicLevel.ERROR, "close: " + t.getMessage(),t);
	        t.printStackTrace();
	        return false;
	    }
	}


	public synchronized byte getMax() {
        return lockLevel;
	}

	private void printDG(StringBuffer sb) {
        Map m = dg.getVertexes();
		if (m.size() > 0) {
	        int ws = waiters.size();
		    sb.append("\ndependency Graph: ");
		    List waiters = new ArrayList(m.keySet());
		    Set s = new HashSet(waiters.size() * 2);
		    s.addAll(m.keySet());
		    for (Iterator it = ((Collection) m.values()).iterator(); it.hasNext();) {
			    s.addAll((Collection) it.next());
		    }			    
		    List all = new ArrayList(s);
		    for (int i = 0; i <all.size() ; i++) {
		        Object t1 = all.get(i);
		        int t1Idx = all.indexOf(t1);
		        Collection dependencies = (Collection) m.get(t1);
		        if (dependencies != null) {
		            for (Iterator it = dependencies.iterator(); it.hasNext();) {
		                sb.append("\nws");
		                sb.append(t1Idx);
		                sb.append(" ");
		                sb.append(t1);
		                sb.append(" = > ");
		                sb.append("ws");
		                Object t2 = it.next();
		                sb.append(all.indexOf(t2));
		                sb.append(" ");
		                sb.append(t2);
		            }
			    }			    
		    }
		}
	}
	
}

final class LockRequest {
	/**
	 * indicates the lock level required by the task
	 */
    private byte lockLevel;

    /**
	 * The task identifier
	 */
    final Object task;
    
	/**
	 * Create a TaskGroup with a single element. If this is a reader group,
	 * additional reader can be added with the #addReader(Object) method. If
	 * this is writer group, the group can not be modified after.
	 * 
	 * @param reader indicates the access mode of the task
	 * @param object is a task (not null)
	 */
    LockRequest(final byte lockLevel, final Object task) {
		this.task = task;
		this.lockLevel = lockLevel;
	}

	final byte getLockLevel() {
		return lockLevel;
	}
	
	Object getTask() {
	    return task;
	}
	
	boolean hasRolledBack() {
	    return false;
	}
	
	public final int hashCode() {
	    return task.hashCode();
	}
	
	public final boolean equals(Object o) {
	    return lockLevel == ((LockRequest) o).lockLevel 
	    	&& task.equals(((LockRequest) o).task);
	}
	
	public final String toString() {
	    return "lockLevel: " + lockLevel + " / tasks: " + task;
	}
}
