/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency;

import junit.framework.TestCase;
import org.objectweb.perseus.persistence.PInteger;
import org.objectweb.perseus.persistence.PIntegerState;
import org.objectweb.perseus.persistence.TPMTest;
import org.objectweb.perseus.cache.replacement.lib.LRUReplacementManager;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.concurrency.POptimisticConcurrencyManager;

public class TestOptimistic extends TestCase {

    PInteger o1, o2;
	Object oid1, oid2;

    TransactionalWorkingSet ctxt1, ctxt2;
	TPMTest tpm;

    public TestOptimistic(String s) throws Exception {
        super(s);
    }

    protected void setUp() throws Exception {

		tpm = new TPMTest("TPM", new POptimisticConcurrencyManager(), new LRUReplacementManager());
		ctxt1 = (TransactionalWorkingSet) tpm.createWS("1");
		ctxt2 = (TransactionalWorkingSet) tpm.createWS("2");
		o1 = new PInteger(1);
		o2 = new PInteger(2);

		tpm.begin(ctxt1);
        oid1 = tpm.export(ctxt1, o1);
		oid2 = tpm.export(ctxt1, o2);
		tpm.prepare(ctxt1);
		tpm.commit(ctxt1);

    }

    // single context, single object, single read access
    public void test1 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.readIntention(ctxt1, oid1);
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
    }

    // single context, single object, single write access
    public void test2 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
    }

    // two contexts, single object, simultaneous read accesses
    public void test3 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, single object, one read access, one write access
    public void test4 () throws ConcurrencyException, PersistenceException {
		PIntegerState s1, s2;
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        s1 = (PIntegerState) tpm.readIntention(ctxt1, oid1);
        s2 = (PIntegerState) tpm.readIntention(ctxt2, oid1);
        s2 = (PIntegerState) tpm.writeIntention(ctxt2, oid1);
		s2.i = 2;

		// check repeatable reads (tx2 not committed)
		s1 = (PIntegerState) tpm.readIntention(ctxt1, oid1);
		assertTrue(s1.i == 1);
        assertTrue(tpm.prepare(ctxt2));
		tpm.commit(ctxt2);

		// check repeatable reads (tx2 committed)
		s1 = (PIntegerState) tpm.readIntention(ctxt1, oid1);
		assertTrue(s1.i == 1);
		// cannot commit read-only transaction with current implementation
        tpm.rollback(ctxt1);
//        tpm.commit(ctxt1);

		// check read committed
		tpm.begin(ctxt1);
		s1 = (PIntegerState) tpm.readIntention(ctxt1, oid1);
		assertTrue(s1.i == 2);
		assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
    }

    // two contexts, single object, one write access, one read access
    public void test5 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
        tpm.rollback(ctxt2);
    }

    // two contexts, single object, simultaneous write access
    public void test6 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        tpm.writeIntention(ctxt2, oid1);
        assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
        tpm.rollback(ctxt2);
    }

    // two contexts, two objects
    // T1 read O1, T2 read O2, T1 write O1, T2 write O2
    public void test7 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid2);
        tpm.writeIntention(ctxt2, oid2);
        assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, two objects, no deadlock with optimist
    // T1 read O1, T2 read O2, T1 write O2, T2 write O1
    public void test9 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid2);
        tpm.readIntention(ctxt1, oid2);
        tpm.writeIntention(ctxt1, oid2);
        tpm.readIntention(ctxt2, oid1);
        tpm.writeIntention(ctxt2, oid1);
        assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
        tpm.rollback(ctxt2);
    }
}
