/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency;

import junit.framework.TestCase;
import org.objectweb.perseus.persistence.PInteger;
import org.objectweb.perseus.persistence.TPMTest;
import org.objectweb.perseus.persistence.PIntegerState;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.replacement.lib.LRUReplacementManager;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.concurrency.PPessimisticConcurrencyManager;

public class TestPessimistic extends TestCase {


	PInteger o1, o2;
	CacheEntry ce1, ce2;
	TransactionalWorkingSet ctxt1, ctxt2, ctxt3;
    Object oid1, oid2;

	TPMTest tpm;



    public TestPessimistic(String s) {
        super(s);
    }

    protected void setUp() throws Exception {

		tpm = new TPMTest("TPM", new PPessimisticConcurrencyManager(), new LRUReplacementManager());
		ctxt1 = (TransactionalWorkingSet) tpm.createWS("1");
		ctxt2 = (TransactionalWorkingSet) tpm.createWS("2");
		ctxt3 = (TransactionalWorkingSet) tpm.createWS("3");
		o1 = new PInteger(1);
		o2 = new PInteger(2);

		tpm.begin(ctxt1);
        oid1 = tpm.export(ctxt1, o1);
		oid2 = tpm.export(ctxt1, o2);
		tpm.prepare(ctxt1);
		tpm.commit(ctxt1);
    }

    // single context, single object, single read access
    public void test1 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.readIntention(ctxt1, oid1);
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
    }

    // single context, single object, single write access
    public void test2 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
    }

    // two contexts, single object, simultaneous read accesses
    public void test3 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        assertTrue(tpm.prepare(ctxt1));
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt1);
        tpm.commit(ctxt2);
    }

    // two contexts, single object, one read access, one write access
    public void test4 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        Thread t = assertLock(new Runnable () {
            public void run () {
                try {
                    tpm.writeIntention(ctxt2, oid1);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
            }
        });
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
        assertUnlock(t);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, single object, one write access, one read access
    public void test5 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        Thread t = assertLock(new Runnable () {
            public void run () {
                try {
                    tpm.readIntention(ctxt2, oid1);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
            }
        });
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
        assertUnlock(t);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, single object, simultaneous write access
    public void test6 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        Thread t = assertLock(new Runnable () {
            public void run () {
                try {
                    tpm.readIntention(ctxt2, oid1);
                    tpm.writeIntention(ctxt2, oid1);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
             }
        });
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
        assertUnlock(t);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, two objects, each context writes its own object (no conflicts)
    public void test7 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.writeIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid2);
        tpm.writeIntention(ctxt2, oid2);
        assertTrue(tpm.prepare(ctxt1));
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt1);
        tpm.commit(ctxt2);
    }

    // two contexts, two objects
    // T1 read O1, T2 read O1, T1 write O2, T2 write O2
    public void test8 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid1);
        tpm.readIntention(ctxt1, oid2);
        tpm.writeIntention(ctxt1, oid2);
        Thread t = assertLock(new Runnable () {
            public void run () {
                try {
                    tpm.readIntention(ctxt2, oid2);
                    tpm.writeIntention(ctxt2, oid2);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
             }
        });
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
        assertUnlock(t);
        assertTrue(tpm.prepare(ctxt2));
        tpm.commit(ctxt2);
    }

    // two contexts, two objects
    // T1 read O1, T2 read O2, T1 write O2, T2 write O1 -> deadlock
    public void test9 () throws ConcurrencyException, PersistenceException {
        tpm.begin(ctxt1);
        tpm.begin(ctxt2);
        tpm.readIntention(ctxt1, oid1);
        tpm.readIntention(ctxt2, oid2);
        Thread t = assertLock(new Runnable () {
            public void run () {
                try {
                    tpm.readIntention(ctxt1, oid2);
                    tpm.writeIntention(ctxt1, oid2);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
             }
        });
        try {
            tpm.readIntention(ctxt2, oid1);
            tpm.writeIntention(ctxt2, oid1);
            fail();
        } catch (PersistenceException _) {
        }
        assertTrue(!tpm.prepare(ctxt2));
        tpm.rollback(ctxt2);

        assertUnlock(t);
        assertTrue(tpm.prepare(ctxt1));
        tpm.commit(ctxt1);
    }

    // asserts that the given action suspends the thread on a lock
    private Thread assertLock (Runnable r) {
        Thread t = new Thread(r);
        t.start();
        try {
            t.join(500);
        } catch (InterruptedException _) {
            // ignored
        }
        assertTrue("thread should be blocked", t.isAlive());
        return t;
    }

    // asserts that the given thread has been unlocked
	private void assertUnlock (Thread t) {
		try {
			t.join(500);
		} catch (InterruptedException _) {
			// ignored
		}
		assertTrue("thread should not be blocked", ! t.isAlive());
	}



// attempt to use closed contexts
	public void test10 () throws ConcurrencyException, PersistenceException {

		ctxt1.setStatus(WorkingSet.CTX_CLOSED);
		/* try to lock prior to opening context */
		try {
			tpm.readIntention(ctxt1, oid1);
			fail("can lock within not open context");
		} catch (PersistenceException _) {
		}
		try {
			tpm.writeIntention(ctxt1, oid1);
			fail("can lock within not open context");
		} catch (PersistenceException _) {
		}
		try {
			tpm.begin(ctxt1);
			fail("can begin tx with closed WS");
		} catch (PersistenceException _) {
		}
		try {
			tpm.prepare(ctxt1);
			fail("can prepare tx with closed WS");
		} catch (PersistenceException _) {
		}
		try {
			tpm.commit(ctxt1);
			fail("can commit tx with closed WS");
		} catch (PersistenceException _) {
		}
		try {
			tpm.rollback(ctxt1);
			fail("can rollback tx with closed WS");
		} catch (PersistenceException _) {
		}
	}


	// verifies that locks are released properly. 2 threads are waiting for a lock (one in read mode,
	// the other in write mode), then their relative contexts are closed.
	// Verifies that when waken up, the threads have gave up their tries and have died.
	// Finally, checks that a new transaction can be granted the lock without waiting.

	public void test11 () throws ConcurrencyException, PersistenceException {

		tpm.begin(ctxt1);
		tpm.begin(ctxt2);
		tpm.begin(ctxt3);

		/* t1 write-locks oid1 */
		tpm.writeIntention(ctxt1, oid1);

		/* t2 blocks in new thread while read-locking oid1 */
		Thread t2 = assertLock(new Runnable () {
			public void run () {
				try {
					tpm.readIntention(ctxt2, oid1);
				} catch (PersistenceException e) {
				}
 			}
		});

		/* t3 blocks in new thread while write-locking oid1 */
		Thread t3 = assertLock(new Runnable () {
			public void run () {
				try {
					tpm.writeIntention(ctxt3, oid1);
				} catch (PersistenceException e) {
				}
 			}
		});


		/* aborts t2 & t3 while they are waiting for a lock */
		tpm.prepare(ctxt2);
		tpm.rollback(ctxt2);
		tpm.prepare(ctxt3);
		tpm.rollback(ctxt3);

		/* checks threads are finished */
		assertUnlock(t2);
		assertUnlock(t3);

		/* validates t1 */
		assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);

		/* checks that new Xact will not block while locking oid1 */
		tpm.begin(ctxt1);
		Thread t4 = new Thread(new Runnable () {
			public void run () {
				try {
					tpm.readIntention(ctxt1, oid1);
				} catch (PersistenceException e) {
					e.printStackTrace();
				}
 			}
		});
		t4.start();
		assertUnlock(t4);

		assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);
	}

	public void test12() throws PersistenceException {
		boolean deadlocked = false;

		tpm.begin(ctxt1);
		tpm.begin(ctxt2);
		Thread t = new Thread(new Runnable () {
			public void run () {
				try {
					tpm.readIntention(ctxt2, oid1);
					tpm.writeIntention(ctxt2, oid1);
				} catch (PersistenceException e) {
					return;
				}
 			}
		});

		PIntegerState s1, s2 ;
		s2 = (PIntegerState) tpm.readIntention(ctxt1, oid2);
		s2  = (PIntegerState) tpm.writeIntention(ctxt1, s2.getCacheEntry());
		s2.i = -1;

		s1 = (PIntegerState) tpm.readIntention(ctxt1, oid1);
		t.start();
		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {
			e.printStackTrace();  //To change body of catch statement use Options | File Templates.
		}
		try {
			tpm.writeIntention(ctxt1, s1.getCacheEntry());
		} catch (PersistenceException e) {
			// should go here
			deadlocked = true;
		}

		if (! deadlocked) fail();
		tpm.rollback(ctxt1);

		tpm.begin(ctxt1);
		s2 = (PIntegerState) tpm.readIntention(ctxt1, oid2);
		if (s2.i == -1) fail();
		assertTrue(tpm.prepare(ctxt1));
		tpm.commit(ctxt1);

	}

}
