/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR AE PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@rd.francetelecom.com)
 *
 */

package org.objectweb.perseus.concurrency.distributed.globallock;


import junit.framework.TestCase;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockManager;
import org.objectweb.perseus.concurrency.distributed.globallock.lib.GlobalLockCoordinatorFactory;
import org.objectweb.perseus.concurrency.distributed.globallock.lib.BasicGlobalLockManager;
import org.objectweb.perseus.distribution.api.DistResControllerService;
import org.objectweb.perseus.distribution.api.DistResServiceManager;
import org.objectweb.perseus.distribution.api.DistResUserService;
import org.objectweb.perseus.distribution.jgroups.JGroupsServiceManager;
import org.objectweb.perseus.distribution.lib.BasicSerializer;
import org.objectweb.perseus.concurrency.distributed.util.RelativeTime;
import org.objectweb.perseus.concurrency.distributed.util.CoordInterceptorFactory;
import org.objectweb.perseus.concurrency.distributed.util.ServiceManagerInterceptor;
import org.objectweb.perseus.concurrency.distributed.util.MessageObserver;
import org.objectweb.perseus.concurrency.distributed.util.MessageEventListenerRegistry;
import org.objectweb.perseus.concurrency.distributed.util.action.Action;
import org.objectweb.perseus.concurrency.distributed.util.action.MoveCoord;
import org.objectweb.perseus.concurrency.distributed.util.action.RunnableActions;
import org.objectweb.perseus.concurrency.distributed.util.action.ThrownException;
import org.objectweb.perseus.concurrency.distributed.util.action.AbstractBlockingAction;
import org.objectweb.perseus.concurrency.distributed.util.event.EventManager;
import org.jgroups.log.Trace;

import java.util.LinkedList;
import java.util.List;
import java.util.Random;


public class TestGlobalLock extends TestCase {

	// number of nodes required and sufficient to run the tests
	static final byte nbNodes = 4;
	// WTU = wait time unit (in milli sec): should be equal to the average
	// time to get a lock without competition. It is used to sequentialized
	// blocking actions and may need to be re-adjusted if the communication
	// latency is too high (in which case, an action could be treated before
	// its preedeceding action)
//	static final long WTU = 400; // 200 seems ok w/ JGroups (400 w/ move coord)
//	static final long standardRunTime = 0 * WTU;
//	static final long deadlockTimeout = 0 * WTU; // 0 = no timeout
	static final long standardRunTime = 0; // 0 = no timeout
	static final long deadlockTimeout = 0; // 0 = no timeout
	static final int nbIterations = 10;

	GlobalLockInterceptor[] h = new GlobalLockInterceptor[nbNodes];
	GlobalLockManagerInterceptor[] glm = new GlobalLockManagerInterceptor[nbNodes];
	String r1 = "r1";
	EventManager em;
	MessageObserver msgObs;
	MessageEventListenerRegistry reg;
	ThrownException te;
	Object[] node = new Object[nbNodes];

	DistResUserService[] drus = new DistResUserService[nbNodes];
	DistResServiceManager drsm;
	CoordInterceptorFactory glcf;

	// moveCoordPeriod values: 0 -> do not move coord, -1 move coord as
	// soon as possible, x>0 -> move coord every x milliseconds
	static final long moveCoordPeriod = 0;
	static final int moveCoordPrintPeriod = 1;
	boolean stopMoveCoord = false;
	Object moveCoordMutex = new Object();
	Thread moveCoordThread;

	List performers;


	// shortcuts
	static final byte P = AbstractBlockingAction.PARALLEL;
	static final byte S = AbstractBlockingAction.SEQUENTIAL;
	static final byte UG = AbstractBlockingAction.GRANTED_BYUSER;
	static final byte CG = AbstractBlockingAction.GRANTED_BYCOORD;
	static final byte UQ = AbstractBlockingAction.QUEUED_BYUSER;
	static final byte CQ = AbstractBlockingAction.QUEUED_BYCOORD;

	static final byte R = RWLockValue.READ;
	static final byte W = RWLockValue.WRITE;
	static final byte N = RWLockValue.NOLOCK;
	static GlobalLockInterceptor GL1, GL2, GL3, GL4;
	static Object N1, N2, N3, N4;

    public TestGlobalLock(String s) throws Exception {
        super(s);
    }

    protected void setUp() throws Exception {
		trace("setting Up:");
		// initialize jgroups traces
		Trace.init();
		msgObs = new MessageObserver();
		reg = new MessageEventListenerRegistry();

		glcf = new CoordInterceptorFactory(
				new GlobalLockCoordinatorFactory(new RWLockValue()),
				msgObs);

		drsm = new ServiceManagerInterceptor(new JGroupsServiceManager(),
				msgObs, reg);

		drsm.setSerializer(new BasicSerializer());

		// Create nodes in reverse order so that the last one is the coordinator
		for (int cpt = nbNodes -1; cpt >= 0; cpt--) {
			trace("Creating node :" + cpt);
			glm[cpt] = new GlobalLockManagerInterceptor(drsm, glcf, reg);
			node[cpt] = glm[cpt].getNodeId();
			drus[cpt] = drsm.getUserService(node[cpt]);
			h[cpt] = (GlobalLockInterceptor) glm[cpt].getGlobalLock(r1, true);
			h[cpt].setName("" + (cpt +1));
		}
		N1 = node[0]; N2 = node[1]; N3 = node[2]; N4 = node[3];
		GL1 = h[0]; GL2 = h[1]; GL3 = h[2]; GL4 = h[3];

		performers = new LinkedList();
		//RelativeTime.wakeUpSkew = (int) WTU / 2; // wake up skew of 50 % of WTU
		em = new EventManager();
		te = new ThrownException();
		AbstractBlockingAction.init(em, te, reg);
    }


	protected void cleanUp()  {
		trace("Cleaning up:");
		if (moveCoordPeriod != 0) {
			trace("Stopping move coord");
			stopMoveCoord = true;
			try {
				moveCoordThread.join(4000);
			} catch (InterruptedException e) {
				moveCoordThread.interrupt();
			}
			trace("Stopping move coord done");
		}

		//TODO: should be merged
		for (int cpt = 0; cpt < nbNodes; cpt++) {
			trace("Stopping node " + cpt);
			((JGroupsServiceManager)
					((ServiceManagerInterceptor) drsm).getIntercepted())
					.stopLocalNode(node[cpt]);
		}
		for (int cpt = 0; cpt < nbNodes; cpt++) {
			trace("Removing node " + cpt);
			drsm.removeLocalNode(node[cpt]);
		}
	}

	private void trace(String s) {
		System.out.println(s);
	}
	private void go() throws Exception {
		go(standardRunTime, nbIterations);
	}
	private void go(long maxTime, int nbIterations) throws Exception {

		em.start();
		te.clearExceptions();

		if (moveCoordPeriod != 0) moveCoordThread = startMoveCoordThread();
		//	Thread.sleep(1000);

		try {
			for (int cpt = 0; cpt < nbIterations; cpt++) {

				trace("TEST #" + cpt);

				List threads = new LinkedList();
				trace("\nExpected:\n" + em.printSchedule(em.expected));

				RelativeTime.startTime(0);
//			Thread.sleep(1000);

				for (int i=0; i < performers.size(); i++) {
					Thread t = new Thread((Runnable) performers.get(i),
							"___ THREAD " + i);
					trace("Start :" + performers.get(i));
					threads.add(t);
					t.start();
				}

				try {
					for (int i=0; i < threads.size(); i++) {

						trace("Wait for thread " + i);
						((Thread) threads.get(i)).join(maxTime);

						if (te.hasExceptions()) {
							trace("Some exceptions have been thrown: " + te);
							throw te;
						}

						if (((Thread) threads.get(i)).isAlive()) {
							trace("Thread " + i + " is blocked");
							// cleanUp();
							em.expected.resetElements();
							trace("\nExpected:\n" + em.printSchedule(em.expected));
							trace("Observed:\n" + em.printSchedule(em.observed));
							fail("There are blocked threads");
						}

					}

					trace("Observed:\n" + em.printSchedule(em.observed));
					trace("END");
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				msgObs.waitEmpty();
				em.reset();
				// uncache locks to have same context at each run
				for (int n = 0; n < nbNodes; n++)
					h[n].uncache();
			//Thread.sleep(1000);
			}
		} finally {
			cleanUp();
		}
	}

	private Thread startMoveCoordThread() {
		Thread t = new Thread(new Runnable() {
			public void run() {
				DistResControllerService drcs0 = drsm.getControllerService(node[0]);
				try {
//					Thread.sleep(1000);
					int numMove = 0;
					Random rand = new Random();
					stopMoveCoord = false;
					while(true) synchronized (moveCoordMutex) {
						if (stopMoveCoord) break;
						int dest;
						do {
							dest = rand.nextInt(nbNodes);
						} while (drcs0.getCoordinators(r1).contains(node[dest]));

						DistResControllerService drcs =
								drsm.getControllerService(node[dest]);
						Action a = new MoveCoord(drcs, r1, em);
						if (numMove % moveCoordPrintPeriod == 0)
							trace("===MOVE=== (" + numMove + ")");
						a.perform();
						if (numMove % moveCoordPrintPeriod == 0)
							trace("===MOVE DONE=== (" + numMove + ")");
						if (moveCoordPeriod > 0) Thread.sleep(moveCoordPeriod);
						numMove++;
					}
					Thread.sleep(1000);
				}  catch (Exception e) {
					e.printStackTrace();
				}
			}
		},  "___MoveCoord");
		t.start();
		return t;
	}

	/**
	 * Test write is not granted while there are conflicts w/ reads
	 * UR1: upgrade to R on node 1 (granted)
	 * UR2: upgrade to R on node 2 (granted)
	 * UR3: upgrade to R on node 3 (granted)
	 * UW4: upgrade to W on node 4 (queued by coord)
	 * DN1: downgrade to N on node 1
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3 (UW4 granted)
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UR1     -->DN1
	 * GL2 (t2)  UR2    |   DN2
	 * GL3 (t3)  UR3    |      -->DN3
	 * GL4 (t4)     -->UW4..........^-->DN4
	 */


	public void testMultipleRBlockOneW() throws Exception {
		trace("testMultipleRBlockOneW");

		Action UR1 = new Upgrade(GL1, CG, S, R, false);
		Action UR2 = new Upgrade(GL2, CG, P, R, false);
		Action UR3 = new Upgrade(GL3, CG, P, R, false);
		Action UW4 = new Upgrade(GL4, CQ, S, W, false);// UW4 QUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		Action DN2 = new Downgrade(GL2, P, N);
		Action DN3 = new Downgrade(GL3, S, N);
		em.addExpectedEvent(UW4.endEvent(), true);            // UW4 UNQUEUED
		Action DN4 = new Downgrade(GL4, S, N);

		List a1 = new LinkedList(); a1.add(UR1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UW4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();

	}

	/**
	 * Test reads are not granted while there are conflicts w/ one write
	 * UW1: upgrade to W on node 1 (granted)
	 * UR2: upgrade to R on node 2 (queued by coord)
	 * UR3: upgrade to R on node 3 (queued by coord)
	 * UR4: upgrade to R on node 4 (queued by coord)
	 * DR1: downgrade to R on node 1 (UR2, UR3, UR4 unqueued)
	 * DN1: downgrade to N on node 1
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UW1    -->DR1-->DN1
	 * GL2 (t2)     -->UR2...!   DN2
	 * GL3 (t3)        UR3...!   DN3
	 * GL4 (t4)        UR4...!   DN4
	 */

	public void testOneWBlocksMultipleR_R() throws Exception {
		trace("testOneWBlocksMultipleR_R");

		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UR2 = new Upgrade(GL2, CQ, S, R, false);// QUEUED
		Action UR3 = new Upgrade(GL3, CQ, P, R, false);// QUEUED
		Action UR4 = new Upgrade(GL4, CQ, P, R, false);// QUEUED
		Action DR1 = new Downgrade(GL1, S, R);
		em.addExpectedEvent(UR2.endEvent(), true);            // UR2 UNQUEUED
		em.addExpectedEvent(UR3.endEvent(), true);            // UR3 UNQUEUED
		em.addExpectedEvent(UR4.endEvent(), true);            // UR4 UNQUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		Action DN2 = new Downgrade(GL2, P, N);
		Action DN3 = new Downgrade(GL3, P, N);
		Action DN4 = new Downgrade(GL4, P, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DR1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UR4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();


	}
	/**
	 * Test reads are not granted while there are conflicts w/ one write
	 * Same as before exept that write is downgraded directly to N
	 * UW1: upgrade to W on node 1 (granted)
	 * UR2: upgrade to R on node 2 (queued by coord)
	 * UR3: upgrade to R on node 3 (queued by coord)
	 * UR4: upgrade to R on node 4 (queued by coord)
	 * DR1: downgrade to N on node 1 (UR2, UR3, UR4 unqueued)
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UW1    -->DRN-->DN1
	 * GL2 (t2)     -->UR2...!   DN2
	 * GL3 (t3)        UR3...!   DN3
	 * GL4 (t4)        UR4...!   DN4
	 */

	public void testOneWBlocksMultipleR_N() throws Exception {
		trace("testOneWBlocksMultipleR_N");

		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UR2 = new Upgrade(GL2, CQ, S, R, false);// QUEUED
		Action UR3 = new Upgrade(GL3, CQ, P, R, false);// QUEUED
		Action UR4 = new Upgrade(GL4, CQ, P, R, false);// QUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UR2.endEvent(), true);            // UR2 UNQUEUED
		em.addExpectedEvent(UR3.endEvent(), true);            // UR3 UNQUEUED
		em.addExpectedEvent(UR4.endEvent(), true);            // UR4 UNQUEUED
		Action DN2 = new Downgrade(GL2, S, N);
		Action DN3 = new Downgrade(GL3, P, N);
		Action DN4 = new Downgrade(GL4, P, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UR4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();



	}
	/**
	 * checks that any W will blocked until current W is downgraded to N
	 * UW1: upgrade to W on node 1 (granted)
	 * UW2: upgrade to W on node 2 (queued by coord)
	 * UW3: upgrade to W on node 3 (queued by coord)
	 * UW4: upgrade to W on node 4 (queued by coord)
	 * DN1: downgrade to N on node 1 (UW2 unqueued)
	 * DN2: downgrade to N on node 2 (UW3 unqueued)
	 * DN3: downgrade to N on node 3 (UW4 unqueued)
	 * GL1 (t1)  UW1             -->DN1
	 * GL2 (t2)     -->UW2......|.....!-->DN2
	 * GL3 (t3)         -->UW3..|...........!-->DN3
	 * GL4 (t4)             -->UW4................!-->DN4
	 */
	public void testOneWBlocksMultipleW() throws Exception {
		trace("testOneWBlocksMultipleW");

		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UW2 = new Upgrade(GL2, CQ, S, W, false);  // QUEUED
		Action UW3 = new Upgrade(GL3, CQ, S, W, false); // QUEUED
		Action UW4 = new Upgrade(GL4, CQ, S, W, false); // QUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2.endEvent(), true);              // UW2 UNQUEUED
		Action DN2 = new Downgrade(GL2, S, N);
		em.addExpectedEvent(UW3.endEvent(), true);              // UW3 UNQUEUED
		Action DN3 = new Downgrade(GL3, S, N);
		em.addExpectedEvent(UW4.endEvent(), true);              // UW4 UNQUEUED
		Action DN4 = new Downgrade(GL4, S, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UW2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UW3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UW4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();


	}
	/**
	 * checks that upgrade to R are blocked if there is already an
	 * pending upgrade to W even though the lock is currently granted in R
	 * UR1: upgrade to R on node 1 (granted)
	 * UW2: upgrade to W on node 2 (queued by coord)
	 * UR3: upgrade to R on node 3 (queued by coord)
	 * UR4: upgrade to R on node 4 (queued by coord)
	 * DN1: downgrade to N on node 1 (UW2 unqueued)
	 * DN2: downgrade to N on node 2 (UR3, UR4 unqueued)
	 * DN3: downgrade to N on node 3
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UR1         -->DN1
	 * GL2 (t2)     -->UW2..|.....!-->DN2
	 * GL3 (t3)         -->UR3..........!-->DN3
	 * GL4 (t4)            UR4..........!   DN4
	 */
	public void test_FIFO_Nodes_R1W2R3R4() throws Exception {
		trace("test_FIFO_Nodes_R1W2R3R4");

		Action UR1 = new Upgrade(GL1, CG, S, R, false);
		Action UW2 = new Upgrade(GL2, CQ, S, W, false);  // QUEUED
		Action UR3 = new Upgrade(GL3, CQ, S, R, false); // QUEUED
		Action UR4 = new Upgrade(GL4, CQ, P, R, false); // QUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2.endEvent(), true);              // UW2 UNQUEUED
		Action DN2 = new Downgrade(GL2, S, N);
		em.addExpectedEvent(UR3.endEvent(), true);              // UR3 UNQUEUED
		em.addExpectedEvent(UR4.endEvent(), true);              // UR4 UNQUEUED
		// then, in parallel
		Action DN3 = new Downgrade(GL3, S, N);
		Action DN4 = new Downgrade(GL4, P, N);

		List a1 = new LinkedList(); a1.add(UR1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UW2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UR4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();
	}



	/**
	 * checks that an upgrade to W is blocked if there is already a
	 * pending upgrade to R and is granted only when all R are downgraded to N
	 * UW1: upgrade to W on node 1 (granted)
	 * UR2: upgrade to R on node 2 (queued by coord)
	 * UR3: upgrade to R on node 3 (queued by coord)
	 * UW4: upgrade to W on node 4 (queued by coord)
	 * DN1: downgrade to N on node 1 (UR2, UR3 unqueued)
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3 (UW4 UNQUEUED)
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UW1          -->DN1
	 * GL2 (t2)     -->UR2...|.....!-->DN2
	 * GL3 (t3)        UR3...|.....!      -->DN3
	 * GL4 (t4)          -->UW4................!-->DN4
	 */

	public void test_FIFO_Nodes_W1R2R3W4() throws Exception {
		trace("test_FIFO_Nodes_W1R2R3W4");
		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UR2 = new Upgrade(GL2, CQ, S, R, false);  // QUEUED
		Action UR3 = new Upgrade(GL3, CQ, P, R, false); // QUEUED
		Action UW4 = new Upgrade(GL4, CQ, S, W, false); // QUEUED
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UR2.endEvent(), true);              // UR2 UNQUEUED
		em.addExpectedEvent(UR3.endEvent(), true);              // UR3 UNQUEUED
		Action DN2 = new Downgrade(GL2, S, N);
		Action DN3 = new Downgrade(GL3, S, N);
		em.addExpectedEvent(UW4.endEvent(), true);              // UW4 UNQUEUED
		Action DN4 = new Downgrade(GL4, S, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UW4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();
	}

	/**
	 * checks that an upgrade to W is blocked if there is already a
	 * pending upgrade to R and is granted only when all R are downgraded to N
	 * UW1: upgrade to W on node 1 (granted)
	 * UR2: upgrade to R on node 2 (queued by coord)
	 * UR3: upgrade to R on node 3 (queued by coord)
	 * UW4: upgrade to W on node 4 (queued by coord)
	 * DN1: downgrade to N on node 1 (UR2, UR3 unqueued)
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3 (UW4 UNQUEUED)
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UW1          -->DR1                  -->DN1
	 * GL2 (t2)     -->UR2...|.....!-->DN2           |
	 * GL3 (t3)        UR3...|.....!      -->DN3     |
	 * GL4 (t4)          -->UW4................!-->DN4
	 */

	public void test_FIFO_Nodes_W1R2R3W4_2() throws Exception {
		trace("test_FIFO_Nodes_W1R2R3W4_2");
		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UR2 = new Upgrade(GL2, CQ, S, R, false);  // QUEUED
		Action UR3 = new Upgrade(GL3, CQ, P, R, false); // QUEUED
		Action UW4 = new Upgrade(GL4, CQ, S, W, false); // QUEUED
		Action DR1 = new Downgrade(GL1, S, R);
		em.addExpectedEvent(UR2.endEvent(), true);              // UR2 UNQUEUED
		em.addExpectedEvent(UR3.endEvent(), true);              // UR3 UNQUEUED
		Action DN2 = new Downgrade(GL2, S, N);
		Action DN3 = new Downgrade(GL3, S, N);
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW4.endEvent(), true);              // UW4 UNQUEUED
		Action DN4 = new Downgrade(GL4, S, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DR1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UW4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();
	}

	/**
     * cheks that W upgrade requests are served in FIFO order and are
	 * not granted until all other nodes have downgraded to N. Also check
	 * that a new arriving R upgrade request will be served after waiting W
	 * UW1: upgrade to W on node 1 (granted)
	 * UW2: upgrade to W on node 2 (queued by coord)
	 * UW3: upgrade to W on node 3 (queued by coord)
	 * UW4: upgrade to W on node 4 (queued by coord)
	 * DR1: downgrade to R on node 1
	 * DN1: downgrade to R on node 1 (UW2 unqueued)
	 * UR1: upgrade to R on node 1 (queued)
	 * DR2: downgrade to R on node 2
	 * DN2: downgrade to N on node 2 (UW3 UNQUEUED)
	 * DR3: downgrade to R on node 3
	 * DN3: downgrade to N on node 3 (UW4 UNQUEUED)
	 * DR4: downgrade to R on node 4 (UR1 UNQUEUED)
	 * DN1b: downgrade to N on node 1
	 * DN4: downgrade to N on node 4
	 * GL1 (t1)  UW1             -->DR1-->DN1-->UR1.......................!-->DN1b
	 * GL2 (t2)     -->UW2......|...........!   DR1-->DN1                 |   |
	 * GL3 (t3)         -->UW3..|.......................!-->DR3-->DN3     |   |
	 * GL4 (t4)             -->UW4..................................!-->DR4   DN4
	 */

	public void test_FIFO_Nodes_W1W2W3W4R1() throws Exception {
		trace("test_FIFO_Nodes_W1W2W3W4R1");

		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UW2 = new Upgrade(GL2, CQ, S, W, false);  // QUEUED
		Action UW3 = new Upgrade(GL3, CQ, S, W, false); // QUEUED
		Action UW4 = new Upgrade(GL4, CQ, S, W, false); // QUEUED
		Action DR1 = new Downgrade(GL1, S, R);
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2.endEvent(), true);              // UW2 UNQUEUED
		Action UR1 = new Upgrade(GL1, CQ, S, R, true);  // QUEUED
		Action DR2 = new Downgrade(GL2, P, R);
		Action DN2 = new Downgrade(GL2, S, N);
		em.addExpectedEvent(UW3.endEvent(), true);              // UW3 UNQUEUED
		Action DR3 = new Downgrade(GL3, S, R);
		Action DN3 = new Downgrade(GL3, S, N);
		em.addExpectedEvent(UW4.endEvent(), true);              // UW4 UNQUEUED
		Action DR4 = new Downgrade(GL4, S, R);
		em.addExpectedEvent(UR1.endEvent(), true);              // UR1 UNQUEUED

		Action DN1b = new Downgrade(GL1, S, N);
		Action DN4 = new Downgrade(GL4, P, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DR1); a1.add(DN1);
		                            a1.add(UR1); a1.add(DN1b);
		List a2 = new LinkedList(); a2.add(UW2); a2.add(DR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UW3); a2.add(DR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(UW4); a2.add(DR4); a4.add(DN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go();
	}


	/* FIFO between threads on 2 different nodes (GL1 & GL2)
	 * Actually, THIS IS NOT REAL FIFO, since R2 will pass while R2 must
	 * wait for W1 to be downgraded what has been called back
	 * Note however that this does not lead to starvation because it will
	 * not prevent R2 from being unblocked
	 *
	 * UR1a: upgrade to R on node 1 (granted)
	 * UW2a: upgrade to W on node 2 (queued because of UR1a)
	 * UR1b: upgrade to R on node 1 (queued after UW2a)
	 * UR2b: upgrade to R on node 2 (queued after UW2a but before UR1b)
	 * DR1a: downgrade to R on node 1
	 * DN1a: downgrade to N on node 1 (UW2a & UR2b unqueued)
	 * DR2: downgrade to R on node 2 (UR1b unqueued)
	 * DN1b: downgrade to N on node 1
	 * DN2b: downgrade to N on node 2
     *
	 * GL1 (t1)  UR1a            -->DR1a-->DN1a
	 * GL1 (t2)     |   -->UR1b.|.............|.......!-->DN1b
	 * GL2 (t3)     -->UW2a.|...|.............!-->DR2a    |
	 * GL2 (t4)             -->UR2b...........!           DN2b
	 *
	 */

	public void test_FIFO_Threads_R1W2R1R2() throws Exception {
		trace("test_FIFO_Threads_R1W2R1R2");
		Action UR1a = new Upgrade(GL1, CG, S, R, false);
		Action UW2a = new Upgrade(GL2, CQ, S, W, false);
		Action UR1b = new Upgrade(GL1, CQ, S, R, true); // must wait Call Back
		Action UR2b = new Upgrade(GL2, UQ, S, R, false);
		Action DR1a = new Downgrade(GL1, S, R);
		Action DN1a = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2a.endEvent(), true);
		em.addExpectedEvent(UR2b.endEvent(), true);
		Action DR2a = new Downgrade(GL2, S, R);
		em.addExpectedEvent(UR1b.endEvent(), true);
		Action DN1b = new Downgrade(GL1, S, N);
		Action DN2b = new Downgrade(GL2, P, N);

		List a1 = new LinkedList(); a1.add(UR1a); a1.add(DR1a); a1.add(DN1a);
		List a2 = new LinkedList(); a2.add(UR1b); a2.add(DN1b);
		List a3 = new LinkedList(); a3.add(UW2a); a3.add(DR2a);
		List a4 = new LinkedList(); a4.add(UR2b); a4.add(DN2b);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL1(t2)", a2, te));
		performers.add(new RunnableActions("GL2(t3)", a3, te));
		performers.add(new RunnableActions("GL2(t4)", a4, te));
		go();


	}

	/* FIFO between threads on 2 different nodes (GL1 & GL2)
   	 * Actually, THIS IS NOT REAL FIFO, since UR2 will be queued before UR1
	 * Note however that this does not lead to starvation because
	 * UR2 will eventually be unqueued
	 *
	 * UW1a: upgrade to W on node 1 (granted)
	 * UW2a: upgrade to W on node 2 (queued because of UW1a)
	 * UR1b: upgrade to R on node 1 (queued after UW2a)
	 * UR2b: upgrade to R on node 2 (queued after UW2a but before UR1b)
	 * DR1a: downgrade to R on node 1
	 * DN1a: downgrade to N on node 1 (UW2a & UR2b unqueued)
	 * DR2a: downgrade to R on node 1 (UR1b unqueued)
	 *
	 * GL1 (t1)  UW1a            -->DR1a-->DN1a
	 * GL1 (t2)     |   -->UR1b.|.............|.......!-->DN1b
	 * GL2 (t3)     -->UW2a.|...|.............!-->DR2a    |
	 * GL2 (t4)             -->UR2b...........!           DN2b
	 *
	 */
	public void test_FIFO_Threads_W1W2R1R2() throws Exception {
		trace("test_FIFO_Threads_W1W2R1R2");

		Action UW1a = new Upgrade(GL1, CG, S, W, false);
		Action UW2a = new Upgrade(GL2, CQ, S, W, false);
		Action UR1b = new Upgrade(GL1, CQ, S, R, true); // must wait Call Back
		Action UR2b = new Upgrade(GL2, UQ, S, R, false);
		Action DR1a = new Downgrade(GL1, S, R);
		Action DN1a = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2a.endEvent(), true);
		em.addExpectedEvent(UR2b.endEvent(), true);
		Action DR2a = new Downgrade(GL2, S, R);
		em.addExpectedEvent(UR1b.endEvent(), true);
		Action DN1b = new Downgrade(GL1, S, N);
		Action DN2b = new Downgrade(GL2, P, N);

		List a1 = new LinkedList(); a1.add(UW1a); a1.add(DR1a); a1.add(DN1a);
		List a2 = new LinkedList(); a2.add(UR1b); a2.add(DN1b);
		List a3 = new LinkedList(); a3.add(UW2a); a3.add(DR2a);
		List a4 = new LinkedList(); a4.add(UR2b); a4.add(DN2b);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL1(t2)", a2, te));
		performers.add(new RunnableActions("GL2(t3)", a3, te));
		performers.add(new RunnableActions("GL2(t4)", a4, te));
		go();



	}

	/* FIFO between threads on 2 different nodes (GL1 & GL2)
	 *
	 * UR1a: upgrade to R on node 1 (granted)
	 * UW2a: upgrade to W on node 2 (queued because of UW1a)
	 * UW1b: upgrade to W on node 1 (queued after UW2a)
	 * UR2b: upgrade to R on node 2 (queued after UW2a but before UW1b)
	 * DN1a: downgrade to N on node 1 (UW2a & UR2b unqueued)
	 * DR2a: downgrade to R on node 1
	 * DN2b: downgrade to N on node 1 (UW1b unqueued)
     * DR1b: downgrade to N on node 1
     *
	 * GL1 (t1)  UR1a            -->DN1a
	 * GL1 (t2)     |   -->UW1b.|......|...........!-->DN1b
	 * GL2 (t3)     -->UW2a.|...|......!-->DR2a    |
	 * GL2 (t4)             -->UR2b....!     -->DN2b
	 *
	 */
	public void test_FIFO_Threads_R1W2W1R2() throws Exception {
		trace("test_FIFO_Threads_R1W2W1R2");

		Action UR1a = new Upgrade(GL1, CG, S, R, false);
		Action UW2a = new Upgrade(GL2, CQ, S, W, false);
		Action UR1b = new Upgrade(GL1, CQ, S, R, true); // must wait call back
		Action UR2b = new Upgrade(GL2, UQ, S, R, false);
		Action DN1a = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UW2a.endEvent(), true);
		em.addExpectedEvent(UR2b.endEvent(), true);
		Action DR2a = new Downgrade(GL2, S, R);
		Action DN2b = new Downgrade(GL2, S, N);
		em.addExpectedEvent(UR1b.endEvent(), true);
		Action DN1b = new Downgrade(GL1, S, N);

		List a1 = new LinkedList(); a1.add(UR1a); a1.add(DN1a);
		List a2 = new LinkedList(); a2.add(UR1b); a2.add(DN1b);
		List a3 = new LinkedList(); a3.add(UW2a); a3.add(DR2a);
		List a4 = new LinkedList(); a4.add(UR2b); a4.add(DN2b);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL1(t3)", a3, te));
		performers.add(new RunnableActions("GL2(t4)", a4, te));
		go();


	}

	/**
	 * Make sure coord failure does not impact lock management
	 * Actually it may impact granting order
	 * UW1: upgrade to W on node 1 (granted)
	 * UW2: upgrade to R on node 2 (queued)
	 * UW3: upgrade to R on node 3 (queued)
	 * KN4 : kill coordinator
	 * DN1: downgrade to N on node 1 (UR2, UR3 unqueued)
	 * DN2: downgrade to N on node 2
	 * DN3: downgrade to N on node 3
	 *
	 * GL1 (t1) UW1          -->DN1
	 * GL2 (t2)   -->UR2....|.....!-->DN2
	 * GL3 (t3)      UR3....|.....!   DN3
	 * GL4 (t4)        -->KN4
	 *
	 */
	public void test_Failure1() throws Exception {
		trace("test_Failure1");
		Action UW1 = new Upgrade(GL1, CG, S, W, false);
		Action UR2 = new Upgrade(GL2, CQ, S, R, false);
		Action UR3 = new Upgrade(GL3, CQ, P, R, false);
		Action KN4 = new KillNodeAndWait("KN4", drsm, N4, r1);
		Action DN1 = new Downgrade(GL1, S, N);
		em.addExpectedEvent(UR2.endEvent(), true);
		em.addExpectedEvent(UR3.endEvent(), true);
		Action DN2 = new Downgrade(GL2, S, N);
		Action DN3 = new Downgrade(GL3, P, N);

		List a1 = new LinkedList(); a1.add(UW1); a1.add(DN1);
		List a2 = new LinkedList(); a2.add(UR2); a2.add(DN2);
		List a3 = new LinkedList(); a3.add(UR3); a3.add(DN3);
		List a4 = new LinkedList(); a4.add(KN4);

		performers.add(new RunnableActions("GL1(t1)", a1, te));
		performers.add(new RunnableActions("GL2(t2)", a2, te));
		performers.add(new RunnableActions("GL3(t3)", a3, te));
		performers.add(new RunnableActions("GL4(t4)", a4, te));
		go(standardRunTime, 1);

	}

}


